from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
import shutil
import gzip
from yucca.paths import yucca_raw_data
from tqdm import tqdm


def convert(path: str, subdir: str = "Hammers"):
    # INPUT DATA
    path = f"{path}/{subdir}"
    file_suffix = ".nii"

    # Train/Test Splits
    labels_dir = join(path, "Labels")
    images_dir = join(path, "Images")

    # S01 to S30 are used as training samples
    # S31 to S40 are used as test samples
    all_samples = subfiles(labels_dir, join=False, suffix=file_suffix)

    assert len(all_samples) > 0

    tr_ids = ["a" + f"{id:02}" for id in range(1, 21)]
    training_samples = [sample for sample in all_samples if sample[: -len(file_suffix)] in tr_ids]
    ts_ids = ["a" + f"{id:02}" for id in range(21, 31)]
    test_samples = [sample for sample in all_samples if sample[: -len(file_suffix)] in ts_ids]

    assert len(training_samples) > 0
    assert len(test_samples) > 0

    # OUTPUT DATA
    # Target names
    task_name = "Task003_Hammers"
    prefix = "Hammers"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    # Populate Target Directory
    # This is likely also the place to apply any re-orientation, resampling and/or label correction.
    for sTr in tqdm(training_samples):
        serial_number = sTr[:-4]
        image_file = open(join(images_dir, sTr), "rb")
        label = open(join(labels_dir, sTr), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTr}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTr}/{prefix}_{serial_number}.nii.gz", "wb"))

    for sTs in tqdm(test_samples):
        serial_number = sTs[:-4]
        image_file = open(join(images_dir, sTs), "rb")
        label = open(join(labels_dir, sTs), "rb")
        shutil.copyfileobj(image_file, gzip.open(f"{target_imagesTs}/{prefix}_{serial_number}_000.nii.gz", "wb"))
        shutil.copyfileobj(label, gzip.open(f"{target_labelsTs}/{prefix}_{serial_number}.nii.gz", "wb"))

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("T1",),
        labels={0: "background", 1: "Left Hippocampus", 2: "Right Hippocampus"},
        dataset_name=task_name,
        license="hands off!",
        dataset_description="Hammers Dataset",
        dataset_reference="",
    )



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "BTCV_Abdomen"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task017_BTCV"
    task_prefix = "BTCV"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    images_dir_tr = join(path, "RawData", "Training", "img")
    labels_dir_tr = join(path, "RawData", "Training", "label")
    images_dir_ts = join(path, "RawData", "Testing", "img")

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for sTr in subfiles(images_dir_tr, suffix=file_suffix, join=False):
        case_id = sTr[: -len(file_suffix)]
        src_image_file_path = join(images_dir_tr, case_id + file_suffix)
        src_label_file_path = join(labels_dir_tr, "label" + case_id[3:] + file_suffix)
        dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{case_id}_000.nii.gz"
        dst_label_file_path = f"{target_labelsTr}/{task_prefix}_{case_id}.nii.gz"

        shutil.copy2(src_image_file_path, dst_image_file_path)
        shutil.copy2(src_label_file_path, dst_label_file_path)

    for sTs in subfiles(images_dir_ts, suffix=file_suffix, join=False):
        case_id = sTs[: -len(file_suffix)]
        src_image_file_path = join(images_dir_ts, case_id + file_suffix)
        dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz"

        shutil.copy2(src_image_file_path, dst_image_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("cine-MRI",),
        labels={
            0: "background",
            1: "spleen",
            2: "right kidney",
            3: "left kidney",
            4: "gallbladder",
            5: "esophagus",
            6: "liver",
            7: "stomach",
            8: "aorta",
            9: "inferior vena cava",
            10: "portal vein and splenic vein",
            11: "pancreas",
            12: "right adrenal gland",
            13: "left adrenal gland",
        },
        dataset_name=task_name,
        license="",
        dataset_description="",
        dataset_reference="",
    )


if __name__ == "__main__":
    convert()



from yucca.pipeline.managers.YuccaManager import YuccaManager


class YuccaManager_PreserveRange(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.augmentation_params["clip_to_input_range"] = True


if __name__ == "__main__":
    # Test that shows the difference between no clipping and clipping to input range.
    # Relevant for models trained on very strict ranges.
    man = YuccaManager_PreserveRange(
        task="Task001_OASIS",
        planner="YuccaPlanner_1_1_1",
        enable_logging=False,
        model_dimensions="2D",
        patch_size=(64, 64),
        batch_size=2,
    )
    man.initialize(stage="fit")
    man.data_module.setup(stage="fit")
    dataset = iter(man.data_module.train_dataloader())
    print("All samples start in the range 0-1. \n Testing the YuccaManager_PreserveRange")
    for i in range(20):
        sample = next(dataset)
        print(f'Sample {i} min value: {sample["image"].min()} and max value: {sample["image"].max()}')



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform


class CopyImageToLabel(YuccaTransform):
    def __init__(self, copy=False, data_key="image", label_key="label"):
        self.copy = copy
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __copy__(self, data_dict):
        data_dict[self.label_key] = data_dict[self.data_key].copy()
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape.\
            \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.data_key].shape}"
        if self.copy:
            data_dict = self.__copy__(data_dict)
        return data_dict



import numpy as np


def motion_ghosting(image, alpha, num_reps, axis, clip_to_input_range):
    img_min = image.min()
    img_max = image.max()
    m = min(0, img_min)
    image += abs(m)
    if len(image.shape) == 3:
        assert axis in [0, 1, 2], "Incorrect or no axis"

        h, w, d = image.shape

        image = np.fft.fftn(image, s=[h, w, d])

        if axis == 0:
            image[0:-1:num_reps, :, :] = alpha * image[0:-1:num_reps, :, :]
        elif axis == 1:
            image[:, 0:-1:num_reps, :] = alpha * image[:, 0:-1:num_reps, :]
        else:
            image[:, :, 0:-1:num_reps] = alpha * image[:, :, 0:-1:num_reps]

        image = abs(np.fft.ifftn(image, s=[h, w, d]))
    if len(image.shape) == 2:
        assert axis in [0, 1], "Incorrect or no axis"
        h, w = image.shape
        image = np.fft.fftn(image, s=[h, w])

        if axis == 0:
            image[0:-1:num_reps, :] = alpha * image[0:-1:num_reps, :]
        else:
            image[:, 0:-1:num_reps] = alpha * image[:, 0:-1:num_reps]
        image = abs(np.fft.ifftn(image, s=[h, w]))
    image -= abs(m)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image



import nibabel as nib
import numpy as np
from PIL import Image
from typing import Union


def np_to_nifti_with_empty_header(array):
    image_file = nib.Nifti1Image(array, affine=None)
    image_file.set_sform(affine=np.eye(4), code=0)
    image_file.set_qform(affine=np.eye(4), code=0)
    return image_file


def nifti_or_np_to_np(array: Union[np.ndarray, nib.Nifti1Image]) -> np.ndarray:
    if isinstance(array, np.ndarray):
        return array
    if isinstance(array, nib.Nifti1Image):
        return array.get_fdata().astype(np.float32)
    else:
        raise TypeError(f"File data type invalid. Found: {type(array)} and expected nib.Nifti1Image or np.ndarray")


def png_to_nifti(path, maybe_to_grayscale=False, is_seg=False, to_label: int = None):
    """Takes a .png and converts it to a Nifti image using Numpy and Nibabel.
    It manually sets both sform and qform codes to 0 (which means they are invalid).
    Therefore, functions relying on trustworthy header information should discard these samples.
    Spacing is set to 1.

    if maybe_to_grayscale = True, it will check if image contains three bands, i.e. if it has color
    channels. If this is the case, it will convert to 'L' which == grayscale
    """
    dtype = np.float32
    if is_seg:
        dtype = np.uint16

    image_file = Image.open(path)
    assert len(image_file.getbands()) in [1, 3], (
        f"unexpected bands in path: {path} " "This can potentially be the alpha band. " "Implement .convert('LA') for this."
    )
    if maybe_to_grayscale and len(image_file.getbands()) == 3:
        image_file = image_file.convert("L")
    image_file = np.array(image_file, dtype=dtype)
    if is_seg and to_label:
        image_file[image_file != 0] = to_label
    image_file = nib.Nifti1Image(image_file, affine=None)
    image_file.set_sform(affine=np.eye(4), code=0)
    image_file.set_qform(affine=np.eye(4), code=0)
    return image_file



import numpy as np
import torch
from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from typing import Optional


class RemoveSegChannelAxis(YuccaTransform):
    def __init__(self, label_key="label", channel_to_remove=1):
        self.label_key = label_key
        self.channel = channel_to_remove

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __remove_channel__(self, image, channel):
        return np.squeeze(image, axis=channel)

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            data_dict[self.label_key].shape[self.channel] == 1
        ), f"Invalid operation: attempting to remove channel of size > 1.\
            \nTrying to squeeze channel: {self.channel} of array with shape: {data_dict[self.label_key].shape}"
        data_dict[self.label_key] = self.__remove_channel__(data_dict[self.label_key], self.channel)
        return data_dict


class NumpyToTorch(YuccaTransform):
    def __init__(self, data_key="image", label_key="label", label_dtype: Optional[torch.dtype] = None):
        self.data_key = data_key
        self.label_key = label_key
        self.label_dtype = label_dtype

    def get_params(self, label):
        if self.label_dtype is not None or label is None:  # Nothing to infer here.
            return
        if isinstance(label, list):  # We just want to look at the first array
            label = label[0]
        if np.issubdtype(label.dtype, (np.floating, float)):
            self.label_dtype = torch.float32
        elif np.issubdtype(label.dtype, int):
            self.label_dtype = torch.int32
        else:
            self.label_dtype = None  # Then we let Torch infer.

    def __convert__(self, datadict):
        datadict[self.data_key] = torch.tensor(datadict[self.data_key], dtype=torch.float32)

        label = datadict.get(self.label_key)

        if label is not None:
            if isinstance(label, list):
                label = [torch.tensor(i, dtype=self.label_dtype) for i in label]
            else:
                label = torch.tensor(label, dtype=self.label_dtype)
            datadict[self.label_key] = label

        return datadict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        data_len = len(data_dict[self.data_key].shape)
        assert (
            data_len == 5 or data_len == 4 or data_len == 3  # (B, C, H, W, D)  # (C, H, W, D) or (B, C, H, W)  # (C, H, W)
        ), f"Incorrect data size or shape.\
            \nShould be (B, C, X, Y, Z) or (B, C, X, Y) or (C, X, Y, Z) or (C, X, Y) and is: {data_len}"
        self.get_params(data_dict.get(self.label_key))
        data_dict = self.__convert__(data_dict)
        return data_dict


class AddBatchDimension(YuccaTransform):
    def __init__(self, data_key="image", label_key="label"):
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        pass

    def __unsqueeze__(self, data_dict):
        data_dict[self.data_key] = data_dict[self.data_key][np.newaxis]
        label = data_dict.get(self.label_key)
        if label is not None:
            if isinstance(label, list):
                label = [s[np.newaxis] for s in label]
            else:
                label = label[np.newaxis]
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        data_dict = self.__unsqueeze__(data_dict)
        return data_dict


class RemoveBatchDimension(YuccaTransform):
    def __init__(self, data_key="image", label_key="label"):
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        pass

    def __squeeze__(self, data_dict):
        data_dict[self.data_key] = data_dict[self.data_key][0]
        label = data_dict.get(self.label_key)
        if label is not None:
            if isinstance(label, list):
                label = [s[0] for s in label]
            else:
                label = label[0]
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        data_dict = self.__squeeze__(data_dict)
        return data_dict


class CollectMetadata(YuccaTransform):
    # Class to collect arbitrary metadata used in later transforms.
    # Values here are collected BEFORE ANY augmentations are applied.
    def __init__(self, data_key="image", label_key="label"):
        self.data_key = data_key
        self.label_key = label_key

    @staticmethod
    def get_params():
        pass

    def __collect__(self, data_dict):
        metadata = {}
        data = data_dict[self.data_key]
        metadata["min"] = data.min()
        metadata["max"] = data.max()
        data_dict["metadata"] = metadata
        return data_dict

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        data_dict = self.__collect__(data_dict)
        return data_dict



import numpy as np


def bias_field(image, clip_to_input_range: bool = False):
    img_min = image.min()
    img_max = image.max()

    if len(image.shape) == 3:
        x, y, z = image.shape
        X, Y, Z = np.meshgrid(
            np.linspace(0, x, x, endpoint=False),
            np.linspace(0, y, y, endpoint=False),
            np.linspace(0, z, z, endpoint=False),
            indexing="ij",
        )
        x0 = np.random.randint(0, x)
        y0 = np.random.randint(0, y)
        z0 = np.random.randint(0, z)
        G = 1 - (np.power((X - x0), 2) / (x**2) + np.power((Y - y0), 2) / (y**2) + np.power((Z - z0), 2) / (z**2))
    else:
        x, y = image.shape
        X, Y = np.meshgrid(
            np.linspace(0, x, x, endpoint=False),
            np.linspace(0, y, y, endpoint=False),
            indexing="ij",
        )
        x0 = np.random.randint(0, x)
        y0 = np.random.randint(0, y)
        G = 1 - (np.power((X - x0), 2) / (x**2) + np.power((Y - y0), 2) / (y**2))
    image = np.multiply(G, image)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image



import numpy as np


def croppad(
    image: np.ndarray,
    image_properties: dict,
    input_dims: np.ndarray,
    patch_size,
    p_oversample_foreground,
    target_image_shape: list | tuple,
    target_label_shape: list | tuple,
    label: np.ndarray = None,
    **pad_kwargs,
):

    if len(patch_size) == 3:
        image, label = croppad_3D_case_from_3D(
            image=image,
            image_properties=image_properties,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 3:
        image, label = croppad_2D_case_from_3D(
            image=image,
            image_properties=image_properties,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
    elif len(patch_size) == 2 and input_dims == 2:
        image, label = croppad_2D_case_from_2D(
            image=image,
            image_properties=image_properties,
            label=label,
            patch_size=patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

    return image, label


def croppad_3D_case_from_3D(
    image, image_properties, label, patch_size, p_oversample_foreground, target_image_shape, target_label_shape, **pad_kwargs
):
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(3):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [(patch_size[d] - image.shape[d + 1])]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2
    pad_lb_z = to_pad[2] // 2
    pad_ub_z = to_pad[2] // 2 + to_pad[2] % 2

    # This is where we should implement any patch selection biases.
    # The final patch excted after augmentation will always be the center of this patch
    # to avoid interpolation artifacts near the borders
    crop_start_idx = []
    if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [np.random.randint(image.shape[d + 1] - patch_size[d] + 1)]
    else:
        location = select_foreground_voxel_to_include(image_properties)
        for d in range(3):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[
        :,
        :,
        :,
        :,
    ] = np.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        **pad_kwargs,
    )
    if label is None:
        return image_out, None
    label_out[
        :,
        :,
        :,
        :,
    ] = np.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
            crop_start_idx[2] : crop_start_idx[2] + patch_size[2],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
    )
    return image_out, label_out


def croppad_2D_case_from_3D(
    image, image_properties, label, patch_size, p_oversample_foreground, target_image_shape, target_label_shape, **pad_kwargs
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 2] < patch_size[d]:
            to_pad += [(patch_size[d] - image.shape[d + 2])]
        else:
            to_pad += [0]

    pad_lb_y = to_pad[0] // 2
    pad_ub_y = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_z = to_pad[1] // 2
    pad_ub_z = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where augmentation-induced interpolation artefacts are least likely
    crop_start_idx = []
    if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= p_oversample_foreground:
        x_idx = np.random.randint(image.shape[1])
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [np.random.randint(image.shape[d + 2] - patch_size[d] + 1)]
    else:
        location = select_foreground_voxel_to_include(image_properties)
        x_idx = location[0]
        for d in range(2):
            if image.shape[d + 2] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d + 1] - patch_size[d]),
                        min(location[d + 1], image.shape[d + 2] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = np.pad(
        image[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
        **pad_kwargs,
    )

    if label is None:
        return image_out, None

    label_out[:, :, :] = np.pad(
        label[
            :,
            x_idx,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_y, pad_ub_y), (pad_lb_z, pad_ub_z)),
    )

    return image_out, label_out


def croppad_2D_case_from_2D(
    image, image_properties, label, patch_size, p_oversample_foreground, target_image_shape, target_label_shape, **pad_kwargs
):
    """
    The possible input for this can be 2D or 3D data.
    For 2D we want to pad or crop as necessary.
    For 3D we want to first select a slice from the first dimension, i.e. volume[idx, :, :],
    then pad or crop as necessary.
    """
    image_out = np.zeros(target_image_shape)
    label_out = np.zeros(target_label_shape)

    # First we pad to ensure min size is met
    to_pad = []
    for d in range(2):
        if image.shape[d + 1] < patch_size[d]:
            to_pad += [(patch_size[d] - image.shape[d + 1])]
        else:
            to_pad += [0]

    pad_lb_x = to_pad[0] // 2
    pad_ub_x = to_pad[0] // 2 + to_pad[0] % 2
    pad_lb_y = to_pad[1] // 2
    pad_ub_y = to_pad[1] // 2 + to_pad[1] % 2

    # This is where we should implement any patch selection biases.
    # The final patch extracted after augmentation will always be the center of this patch
    # as this is where artefacts are least present
    crop_start_idx = []
    if len(image_properties["foreground_locations"]) == 0 or np.random.uniform() >= p_oversample_foreground:
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [np.random.randint(image.shape[d + 1] - patch_size[d] + 1)]
    else:
        location = select_foreground_voxel_to_include(image_properties)
        for d in range(2):
            if image.shape[d + 1] < patch_size[d]:
                crop_start_idx += [0]
            else:
                crop_start_idx += [
                    np.random.randint(
                        max(0, location[d] - patch_size[d]),
                        min(location[d], image.shape[d + 1] - patch_size[d]) + 1,
                    )
                ]

    image_out[:, :, :] = np.pad(
        image[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
        **pad_kwargs,
    )

    if label is None:  # Reconstruction/inpainting
        return image_out, None

    if len(label.shape) == 1:  # Classification
        return image_out, label

    label_out[:, :, :] = np.pad(
        label[
            :,
            crop_start_idx[0] : crop_start_idx[0] + patch_size[0],
            crop_start_idx[1] : crop_start_idx[1] + patch_size[1],
        ],
        ((0, 0), (pad_lb_x, pad_ub_x), (pad_lb_y, pad_ub_y)),
    )

    return image_out, label_out


def select_foreground_voxel_to_include(image_properties):
    if isinstance(image_properties["foreground_locations"], list):
        locidx = np.random.choice(len(image_properties["foreground_locations"]))
        location = image_properties["foreground_locations"][locidx]
    elif isinstance(image_properties["foreground_locations"], dict):
        selected_class = np.random.choice(list(image_properties["foreground_locations"].keys()))
        locidx = np.random.choice(len(image_properties["foreground_locations"][selected_class]))
        location = image_properties["foreground_locations"][selected_class][locidx]
    return location



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
import numpy as np
from yucca.functional.transforms import bias_field


class BiasField(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        p_per_sample: float = 1.0,
        clip_to_input_range=False,
    ):
        self.data_key = data_key
        self.p_per_sample = p_per_sample
        self.clip_to_input_range = clip_to_input_range

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __biasField__(self, image):
        image = bias_field(image, clip_to_input_range=self.clip_to_input_range)
        return image

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        assert (
            len(data_dict[self.data_key].shape) == 5 or len(data_dict[self.data_key].shape) == 4
        ), f"Incorrect data size or shape. \nShould be (b, c, x, y, z) or (b, c, x, y) and is:\
                {data_dict[self.data_key].shape}"

        for b in range(data_dict[self.data_key].shape[0]):
            for c in range(data_dict[self.data_key][b].shape[0]):
                if np.random.uniform() < self.p_per_sample:
                    data_dict[self.data_key][b, c] = self.__biasField__(data_dict[self.data_key][b, c])
        return data_dict



"""
Cross Entropy and Dice Loss from the paper:
Isensee, F., Jaeger, P.F., Kohl, S.A.A. et al.
    nnU-Net: a self-configuring method for deep learning-based biomedical image segmentation.
    Nat Methods 18, 203–211 (2021). https://doi.org/10.1038/s41592-020-01008-z

Implementation by:
    https://github.com/Sllambias/UNetDA/blob/main/nnUNet/nnunet/training/loss_functions/dice_loss.py
"""

import torch
import numpy as np
from torch import nn
from yucca.optimization.loss_functions.CE import CE
import torch.nn.functional as F


def sum_tensor(inp, axes, keepdim=False):
    axes = np.unique(axes).astype(int)
    if keepdim:
        for ax in axes:
            inp = inp.sum(int(ax), keepdim=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp


def get_tp_fp_fn_tn(net_output, gt, axes=None, mask=None, square=False):
    """
    net_output must be (b, c, x, y(, z)))
    gt must be a label map (shape (b, 1, x, y(, z)) OR shape (b, x, y(, z)))
        or one hot encoding (b, c, x, y(, z))
    if mask is provided it must have shape (b, 1, x, y(, z)))
    :param net_output:
    :param gt:
    :param axes: can be (, ) = no summation
    :param mask: mask must be 1 for valid pixels and 0 for invalid pixels
    :param square: if True then fp, tp and fn will be squared before summation
    :return:
    """

    if axes is None:
        axes = tuple(range(2, len(net_output.size())))

    shp_x = net_output.shape
    shp_y = gt.shape
    with torch.no_grad():
        if len(shp_x) != len(shp_y):
            gt = gt.view((shp_y[0], 1, *shp_y[1:]))

        if all([i == j for i, j in zip(net_output.shape, gt.shape)]):
            # if this is the case then gt is probably already a one hot encoding
            y_onehot = gt
        else:
            gt = gt.long()
            y_onehot = torch.zeros(shp_x).to(net_output)
            y_onehot.scatter_(1, gt, 1)

    tp = net_output * y_onehot
    fp = net_output * (1 - y_onehot)
    fn = (1 - net_output) * y_onehot
    tn = (1 - net_output) * (1 - y_onehot)

    if mask is not None:
        tp = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(tp, dim=1)), dim=1)
        fp = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(fp, dim=1)), dim=1)
        fn = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(fn, dim=1)), dim=1)
        tn = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(tn, dim=1)), dim=1)

    if square:
        tp = tp**2
        fp = fp**2
        fn = fn**2
        tn = tn**2

    if len(axes) > 0:
        tp = sum_tensor(tp, axes, keepdim=False)
        fp = sum_tensor(fp, axes, keepdim=False)
        fn = sum_tensor(fn, axes, keepdim=False)
        tn = sum_tensor(tn, axes, keepdim=False)

    return tp, fp, fn, tn


class SoftDiceLoss(nn.Module):
    def __init__(self, apply_softmax=True, batch_dice=False, do_bg=True, smooth=1.0):
        """ """
        super(SoftDiceLoss, self).__init__()

        self.do_bg = do_bg
        self.batch_dice = batch_dice
        self.smooth = smooth
        self.apply_softmax = apply_softmax

    def forward(self, x, y, loss_mask=None):
        shp_x = x.shape

        if self.batch_dice:
            axes = [0] + list(range(2, len(shp_x)))
        else:
            axes = list(range(2, len(shp_x)))

        if self.apply_softmax is True:
            x = F.softmax(x, 1)

        tp, fp, fn, _ = get_tp_fp_fn_tn(x, y, axes, loss_mask, False)

        nominator = 2 * tp + self.smooth
        denominator = 2 * tp + fp + fn + self.smooth

        dc = nominator / (denominator + 1e-8)

        if not self.do_bg:
            if self.batch_dice:
                dc = dc[1:]
            else:
                dc = dc[:, 1:]
        dc = dc.mean()

        return -dc


class DiceCE(nn.Module):
    def __init__(
        self,
        soft_dice_kwargs={},
        ce_kwargs={},
        weight_ce=1,
        weight_dice=1,
        log_dice=False,
        ignore_label=None,
    ):
        """
        CAREFUL. Weights for CE and Dice do not need to sum to one. You can set whatever you want.
        :param soft_dice_kwargs:
        :param ce_kwargs:
        :param aggregate:
        :param square_dice:
        :param weight_ce:
        :param weight_dice:
        """
        super(DiceCE, self).__init__()
        if ignore_label is not None:
            ce_kwargs["reduction"] = "none"
        self.log_dice = log_dice
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce
        self.ce = CE()

        self.ignore_label = ignore_label

        self.dc = SoftDiceLoss(**soft_dice_kwargs)

    def forward(self, net_output, target):
        """
        target must be b, c, x, y(, z) with c=1
        :param net_output:
        :param target:
        :return:
        """
        if self.ignore_label is not None:
            assert target.shape[1] == 1, "not implemented for one hot encoding"
            mask = target != self.ignore_label
            target[~mask] = 0
            mask = mask.float()
        else:
            mask = None

        dc_loss = self.dc(net_output, target, loss_mask=mask) if self.weight_dice != 0 else 0
        if self.log_dice:
            dc_loss = -torch.log(-dc_loss)

        ce_loss = self.ce(net_output, target[:, 0].long()) if self.weight_ce != 0 else 0
        if self.ignore_label is not None:
            ce_loss *= mask[:, 0]
            ce_loss = ce_loss.sum() / mask.sum()

        result = self.weight_ce * ce_loss + self.weight_dice * dc_loss

        return result



import numpy as np
from typing import List
from yucca.functional.transforms.label_transforms import convert_labels_to_regions
from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform


class ConvertLabelsToRegions(YuccaTransform):
    def __init__(self, convert_labels_to_regions=False, label_key="label", regions: List[List[int]] = None):
        self.convert_labels_to_regions = convert_labels_to_regions
        self.label_key = label_key
        self.regions = regions
        if self.convert_labels_to_regions:
            assert self.regions is not None, "you cannot enable convert_labels_to_regions while not supplying any regions"

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __convert__(self, label: np.ndarray, regions: List[List[int]]):
        return convert_labels_to_regions(label, regions)

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        if self.convert_labels_to_regions:
            assert (
                len(data_dict[self.label_key].shape) == 5 or len(data_dict[self.label_key].shape) == 4
            ), f"Incorrect data size or shape.\
                \nShould be (b, c, x, y, z) or (b, c, x, y) and is: {data_dict[self.label_key].shape}"
            data_dict[self.label_key] = self.__convert__(data_dict[self.label_key], self.regions)
        return data_dict



import pkgutil
import importlib
import fileinput
import re
import shutil
import os
from batchgenerators.utilities.file_and_folder_operations import (
    join,
    subdirs,
)
from typing import Union, List


def replace_in_file(file_path, pattern_replacement):
    with fileinput.input(file_path, inplace=True) as file:
        for line in file:
            for pattern, replacement in pattern_replacement.items():
                line = line.replace(pattern, replacement)
            print(line, end="")


def subfiles(
    folder: str,
    join: bool = True,
    remove_extension: bool = False,
    prefix: Union[List[str], str] = None,
    suffix: Union[List[str], str] = None,
    sort: bool = True,
) -> List[str]:
    # Copy of batchgenerators subfiles with the added "remove_extension" arg

    if join:
        l = os.path.join
    else:
        l = lambda x, y: y  # noqa: U100

    if remove_extension:
        r = lambda x: x.split(".")[0]  # regular os.splitext doesn't account for double extensions like .nii.gz or .tar.gz
    else:
        r = lambda x: x

    if prefix is not None and isinstance(prefix, str):
        prefix = [prefix]
    if suffix is not None and isinstance(suffix, str):
        suffix = [suffix]

    res = [
        r(l(folder, i))
        for i in os.listdir(folder)
        if os.path.isfile(os.path.join(folder, i))
        and (prefix is None or any([i.startswith(j) for j in prefix]))
        and (suffix is None or any([i.endswith(j) for j in suffix]))
    ]

    if sort:
        res.sort()
    return res


def rename_file_or_dir(file: str, patterns: dict):
    # Patterns is a dict of key, value pairs where keys are the words to replace and values are
    # what to substitute them by. E.g. if patterns = {"foo": "bar"}
    # then the sentence "foo bar" --> "bar bar"
    newfile = file
    for k, v in patterns.items():
        newfile = re.sub(k, v, newfile)
    if os.path.isdir(file):
        if newfile != file:
            shutil.move(file, newfile)
    elif os.path.isfile(file):
        os.rename(file, newfile)
    return newfile


def recursive_rename(folder, patterns_in_file={}, patterns_in_name={}, file_ext=".py"):
    """
    Takes a top folder and recursively looks through all subfolders and files.
    For all file contents it will replace patterns_in_file keys with the corresponding values.
    For all file names it will replace the patterns_in_name keys with the corresponding values.

    If patterns_in_file = {"llama": "alpaca", "coffee": "TEA"}
    and patterns_in_name = {"foo": "bar", "Foo": "Bar"}
    and we take the file:

    MyPythonFooScript.py
    ---- (with the following lines of code) ----
    llama = 42
    coffee = 123

    something_else = llama + coffee
    ----

    then we will end up with

    MyPythonBarScript.py
    ----
    alpaca = 42
    TEA = 123

    something_else = alpaca + TEA
    """
    dirs = subdirs(folder)
    files = subfiles(folder)
    for file in files:
        if len(patterns_in_file) > 0:
            if os.path.splitext(file)[-1] == file_ext:
                replace_in_file(
                    file,
                    patterns_in_file,
                )
        rename_file_or_dir(file, patterns_in_name)
    for direc in dirs:
        direc = rename_file_or_dir(direc, patterns_in_name)
        recursive_rename(direc, patterns_in_file, patterns_in_name)


def recursive_find_python_class(folder: list, class_name: str, current_module: str):
    """
    Stolen from nnUNet model_restore.py.
    Folder = starting path, e.g. join(yucca.__path__[0], 'preprocessing')
    Trainer_name = e.g. YuccaPreprocessor3D
    Current_module = starting module e.g. 'yucca.preprocessing'

    The function is nested to allow it to exit after the recursive part is done. This lets us to
    check if we have successfully found what we looked for IN the function.
    """

    def _recursive_find_python_class(folder: list, class_name: str, current_module: str):
        tr = None
        for _, modname, ispkg in pkgutil.iter_modules(folder):
            if not ispkg:
                m = importlib.import_module(current_module + "." + modname)
                if hasattr(m, class_name):
                    tr = getattr(m, class_name)
                    break

        if tr is None:
            for _, modname, ispkg in pkgutil.iter_modules(folder):
                if ispkg:
                    next_current_module = current_module + "." + modname
                    tr = _recursive_find_python_class(
                        [join(folder[0], modname)],
                        class_name,
                        current_module=next_current_module,
                    )
                if tr is not None:
                    break
        return tr

    found_python_class = _recursive_find_python_class(folder, class_name, current_module)
    assert (
        found_python_class is not None
    ), f"Did not find any python class called {class_name}. Make sure there's no typos (and that the class actually exists)"
    return found_python_class


def recursive_find_realpath(path):
    """
    This might produce undesirable results if run on a slurm/batch management user, that does not
    the same permissions as you.
    """
    non_linked_dirs = []
    while path:
        if os.path.islink(path):
            path = os.path.realpath(path)
        path, part = os.path.split(path)
        non_linked_dirs.append(part)
        if path == os.path.sep:
            non_linked_dirs.append(path)
            path = False
    return os.path.join(*non_linked_dirs[::-1])



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json, should_use_volume
from yucca.paths import yucca_raw_data
from tqdm import tqdm

import nibabel as nib


def convert(path: str, subdir: str = "decathlon/Task04_Hippocampus"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    ext = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task205_Hippocampus"
    task_prefix = "Hippocampus"

    """ Access the input data. If images are not split into train/test, and you wish to randomly
    split the data, uncomment and adapt the following lines to fit your local path. """

    subjects_dir = join(path, "imagesTr")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    skipped_volumes = []

    for file in tqdm(subfiles(subjects_dir, join=False, suffix=ext)):
        image_path = join(subjects_dir, file)
        file_name = file[: -len(ext)]
        vol = nib.load(image_path)
        if should_use_volume(vol):
            output_name = f"{task_prefix}_{file_name}_000.nii.gz"
            output_path = join(target_imagesTr, output_name)
            shutil.copy2(image_path, output_path)
        else:
            skipped_volumes.append(image_path)

    print(f"Skipped {len(skipped_volumes)}")
    print("Skipped volumes", skipped_volumes)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels=None,
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="Decathlon: Left and right hippocampus segmentation",
        dataset_reference="Vanderbilt University Medical Center",
    )



# ONLY used to run tests
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
import shutil
from yucca.paths import yucca_raw_data
from tqdm import tqdm
from sklearn.model_selection import train_test_split


def convert(path: str, subdir: str = "dataset_test0_classification"):
    # INPUT DATA
    path = join(path, subdir)
    label_suffix = ".txt"
    image_suffix = ".nii.gz"

    # Train/Test Splits
    labels_dir = join(path, "all_labels")
    images_dir = join(path, "all_images")

    training_samples, test_samples = train_test_split(
        subfiles(labels_dir, join=False, suffix=label_suffix), random_state=859032
    )

    # OUTPUT DATA
    # Target names
    task_name = "Task000_TEST_CLASSIFICATION"
    task_prefix = "TEST_CLASSIFICATION"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    for sTr in tqdm(training_samples, desc="Train"):
        sTr = sTr[: -len(label_suffix)]
        src_image_file_path = join(images_dir, sTr + image_suffix)
        src_label_path = join(labels_dir, sTr + label_suffix)

        dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"
        dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr}.txt"
        shutil.copy2(src_image_file_path, dst_image_file_path)
        shutil.copy2(src_label_path, dst_label_path)
    del sTr

    for sTs in tqdm(test_samples, desc="Test"):
        sTs = sTs[: -len(label_suffix)]

        # Loading relevant modalities and the ground truth
        src_image_file_path = join(images_dir, sTs + image_suffix)
        src_label_path = join(labels_dir, sTs + label_suffix)

        dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"
        dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"
        shutil.copy2(src_image_file_path, dst_image_file_path)
        shutil.copy2(src_label_path, dst_label_path)
    del sTs

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        ("TestModality",),
        labels={0: "background", 1: "FakeLabel1"},
        dataset_name=task_name,
        dataset_description="Test Dataset",
        dataset_reference="",
    )


# %%



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfolders
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data


def convert(path: str, subdir: str = "WMH"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task208_WMH"
    task_prefix = "WMH"

    datasets = ["Amsterdam", "Singapore", "Utrecht"]
    # site = ""

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    ###Populate Target Directory###
    for dataset in datasets:
        dataset_path = join(path, dataset)
        if dataset == "Amsterdam":
            tr_folder = "Train_GE3T"
        else:
            tr_folder = "Train"

        train_folder = join(dataset_path, tr_folder)
        # test_folder = join(dataset_path, "Test")

        training_samples = subfolders(train_folder, join=False)
        # test_samples = subfolders(test_folder, join=False)

        # First we add the data from the training folders
        for sTr in training_samples:
            # Loading relevant modalities and the ground truth
            src_flair_file_path = join(train_folder, sTr, "pre", "FLAIR.nii.gz")
            src_t1_file_path = join(train_folder, sTr, "pre", "T1.nii.gz")
            dst_flair_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_flair_000.nii.gz"
            dst_t1_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_t1_000.nii.gz"
            shutil.copy2(src_flair_file_path, dst_flair_file_path)
            shutil.copy2(src_t1_file_path, dst_t1_file_path)

        # Then we add the data from the original testing folders
        # if dataset == "Amsterdam":
        #   for site in test_samples:
        #       samples = subfolders(join(test_folder, site), join=False)
        #       for sTr in samples:
        #           # Loading relevant modalities and the ground truth
        #           src_flair_file_path = join(test_folder, site, sTr, "pre", "FLAIR.nii.gz")
        #           src_t1_file_path = join(test_folder, site, sTr, "pre", "T1.nii.gz")
        #           dst_flair_file_path = f"{target_imagesTr}/{task_prefix}_flair_{sTr}_000.nii.gz"
        #           dst_t1_file_path = f"{target_imagesTr}/{task_prefix}_t1_{sTr}_000.nii.gz"
        #           shutil.copy2(src_flair_file_path, dst_flair_file_path)
        #           shutil.copy2(src_t1_file_path, dst_t1_file_path)
    #
    # else:
    #   for sTr in test_samples:
    #       # Loading relevant modalities and the ground truth
    #       src_flair_file_path = join(test_folder, sTr, "pre", "FLAIR.nii.gz")
    #       src_t1_file_path = join(test_folder, sTr, "pre", "T1.nii.gz")
    #       dst_flair_file_path = f"{target_imagesTr}/{task_prefix}_flair_{sTr}_000.nii.gz"
    #       dst_t1_file_path = f"{target_imagesTr}/{task_prefix}_t1_{sTr}_000.nii.gz"
    #       shutil.copy2(src_flair_file_path, dst_flair_file_path)
    #       shutil.copy2(src_t1_file_path, dst_t1_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=("Flair/T1",),
        labels={},
        dataset_name=task_name,
        license="CC BY-NC 4.0 DEED",
        dataset_description="White Matter Hyperintensity Segmentation Challenge. Preprocessed Flair and T1 images and NO labels!",
        dataset_reference="https://dataverse.nl/dataset.xhtml?persistentId=doi:10.34894/AECRSD",
    )



import numpy as np
from skimage.transform import resize


def downsample_label(label, factors):
    orig_type = label.dtype
    orig_shape = label.shape
    downsampled_labels = []
    for factor in factors:
        target_shape = np.array(orig_shape).astype(int)
        for i in range(2, len(orig_shape)):
            target_shape[i] *= factor
        if np.all(target_shape == orig_shape):
            downsampled_labels.append(label)
        else:
            canvas = np.zeros(target_shape)
            for b in range(label.shape[0]):
                for c in range(label[b].shape[0]):
                    canvas[b, c] = resize(
                        label[b, c].astype(float),
                        target_shape[2:],
                        0,
                        mode="edge",
                        clip=True,
                        anti_aliasing=False,
                    ).astype(orig_type)
            downsampled_labels.append(canvas)
    return downsampled_labels


def simulate_lowres(image, target_shape, clip_to_input_range):
    img_min = image.min()
    img_max = image.max()
    shape = image.shape
    downsampled = resize(
        image.astype(float),
        target_shape,
        order=0,
        mode="edge",
        anti_aliasing=False,
    )
    image = resize(downsampled, shape, order=3, mode="edge", anti_aliasing=False)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image



import nibabel as nib
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subdirs
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source
from yucca.functional.utils.nib_utils import reorient_to_RAS


def convert(path: str = yucca_source, subdir: str = "kits23"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)
    file_suffix = ".nii.gz"

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task018_KITS23"
    task_prefix = "KITS23"

    """ Access the input data. If images are not split into train/test, and you wish to randomly 
    split the data, uncomment and adapt the following lines to fit your local path. """

    images_dir_tr = labels_dir_tr = join(path, "dataset")

    """ Then define target paths """
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    """Populate Target Directory
    This is also the place to apply any re-orientation, resampling and/or label correction."""

    for sTr in subdirs(images_dir_tr, join=False):
        src_image_file_path = join(images_dir_tr, sTr, "imaging" + file_suffix)
        src_label_file_path = join(labels_dir_tr, sTr, "segmentation" + file_suffix)

        dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"
        dst_label_file_path = f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz"

        image = nib.load(src_image_file_path)
        label = nib.load(src_label_file_path)

        image = reorient_to_RAS(image)
        label = reorient_to_RAS(label)

        nib.save(image, dst_image_file_path)
        nib.save(label, dst_label_file_path)

    # for sTs in subfiles(images_dir_ts, suffix=file_suffix, join=False):
    #    case_id = sTs[: -len(file_suffix)]
    #    src_image_file_path = join(images_dir_ts, case_id + file_suffix)
    #    dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{case_id}_000.nii.gz"
    #    shutil.copy2(src_image_file_path, dst_image_file_path)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("CT",),
        labels={0: "background", 1: "kidney", 2: "tumor", 3: "cyst"},
        regions_in_order=[[1, 2, 3], [2, 3], [2]],
        regions_labeled=[1, 3, 2],
        dataset_name=task_name,
        license="",
        dataset_description="The 2023 Kidney and Kidney Tumor Segmentation Challenge",
        dataset_reference="https://kits-challenge.org/kits23",
    )


if __name__ == "__main__":
    convert()



from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms import downsample_label


class DownsampleSegForDS(YuccaTransform):
    """ """

    def __init__(self, deep_supervision: bool = False, label_key="label", factors=(1, 0.5, 0.25, 0.125, 0.0625)):
        self.deep_supervision = deep_supervision
        self.label_key = label_key
        self.factors = factors

    @staticmethod
    def get_params():
        # No parameters to retrieve
        pass

    def __downsample__(self, label, factors):
        downsampled_labels = downsample_label(label, factors)
        return downsampled_labels

    def __call__(self, packed_data_dict=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict
        if self.deep_supervision:
            data_dict[self.label_key] = self.__downsample__(data_dict[self.label_key], self.factors)
        return data_dict



import os
from yucca.pipeline.task_conversion.utils import combine_imagesTr_from_tasks, generate_dataset_json
from yucca.paths import yucca_raw_data


def convert(_path: str, _subdir: str = ""):
    # Define the name of the new task
    task_name = "Task100_combine_001_002"

    # Define the expected labels (leave empty for no labels)
    expected_labels = {}

    # Define expected modalities, such as ('MRI', 'CT',) or just ('Medical Images',) - remember trailing "," for singular modalities
    expected_modalities = ("MRI",)

    # Define the tasks to combine, such as ["Task001_OASIS", "Task002_LPBA40"]
    # The individual task_conversion scripts must be run prior to executing this, as the script will look for the data in the yucca_raw_data folder.
    tasks_to_combine = ["Task001_OASIS", "Task002_LPBA40"]

    ### In most cases the remaining can be left untouched ###
    # Setting the paths to save the new task and making the directories
    target_base = os.path.join(yucca_raw_data, task_name)
    target_imagesTr = os.path.join(yucca_raw_data, task_name, "imagesTr")
    target_imagesTs = None
    os.makedirs(target_imagesTr, exist_ok=True)

    combine_imagesTr_from_tasks(tasks=tasks_to_combine, target_dir=target_imagesTr)

    generate_dataset_json(
        os.path.join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=expected_modalities,
        labels=expected_labels,
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="",
        dataset_reference="",
    )



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner


class YuccaPlanner_Compress(YuccaPlanner):
    def __init__(self, task, preprocessor=None, threads=None, disable_sanity_checks=False, view=None):
        super().__init__(
            task,
            preprocessor=preprocessor,
            threads=threads,
            disable_sanity_checks=disable_sanity_checks,
            view=view,
        )
        self.name = str(self.__class__.__name__)
        self.compress = True



import numpy as np


def gibbs_ringing(image, num_sample, axis, clip_to_input_range: bool = False):
    img_min = image.min()
    img_max = image.max()
    m = min(0, img_min)
    image += abs(m)
    if len(image.shape) == 3:
        assert axis in [0, 1, 2], "Incorrect or no axis"

        h, w, d = image.shape
        if axis == 0:
            image = image.transpose(0, 2, 1)
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, d, w]))
            image[:, :, 0 : int(np.ceil(w / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(w / 2) + np.ceil(num_sample / 2)) : w] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, d, w]))
            image = image.transpose(0, 2, 1)
        elif axis == 1:
            image = image.transpose(1, 2, 0)
            image = np.fft.fftshift(np.fft.fftn(image, s=[w, d, h]))
            image[:, :, 0 : int(np.ceil(h / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(h / 2) + np.ceil(num_sample / 2)) : h] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[w, d, h]))
            image = image.transpose(2, 0, 1)
        else:
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, w, d]))
            image[:, :, 0 : int(np.ceil(d / 2) - np.ceil(num_sample / 2))] = 0
            image[:, :, int(np.ceil(d / 2) + np.ceil(num_sample / 2)) : d] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, w, d]))
    elif len(image.shape) == 2:
        assert axis in [0, 1], "incorrect or no axis"
        h, w = image.shape
        if axis == 0:
            image = np.fft.fftshift(np.fft.fftn(image, s=[h, w]))
            image[:, 0 : int(np.ceil(w / 2) - np.ceil(num_sample / 2))] = 0
            image[:, int(np.ceil(w / 2) + np.ceil(num_sample / 2)) : w] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[h, w]))
        else:
            image = image.conj().T
            image = np.fft.fftshift(np.fft.fftn(image, s=[w, h]))
            image[:, 0 : int(np.ceil(h / 2) - np.ceil(num_sample / 2))] = 0
            image[:, int(np.ceil(h / 2) + np.ceil(num_sample / 2)) : h] = 0
            image = abs(np.fft.ifftn(np.fft.ifftshift(image), s=[w, h]))
            image = image.conj().T
    image -= abs(m)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image



from yucca.data.datasets.YuccaDataset import YuccaTrainDataset
import numpy as np


class YuccaTrainDataset_1modality(YuccaTrainDataset):
    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super().__init__(
            *args,
            **kwargs,
        )

    def unpack(self, data, supervised: bool):
        images, label = super().unpack(data, supervised)
        chosen_modality = np.random.randint(0, images.shape[0])

        # by slicing this way we keep the channel dimension
        # rather than doing images[chosen_modality] which makes (c,h,w,d) --> (h,w,d)
        images = images[chosen_modality : chosen_modality + 1]
        return images, label

    def unpack_with_zeros(self, data, supervised: bool):
        images, label = super().unpack_with_zeros(data, supervised)

        # first find a modality that is not a zero-array
        valid_modalities = [i for i in range(images.shape[0]) if np.any(data[i])]

        chosen_modality = np.random.choice(valid_modalities)
        images = images[chosen_modality : chosen_modality + 1]
        return images, label



from yucca.pipeline.preprocessing import ClassificationPreprocessor
import yucca
import numpy as np
from yucca.paths import yucca_preprocessed_data, yucca_raw_data
from yucca.pipeline.planning.dataset_properties import create_dataset_properties
from yucca.functional.utils.files_and_folders import recursive_find_python_class
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, isfile, load_pickle, save_json
from yucca.functional.planning import make_plans_file, add_stats_to_plans_post_preprocessing
from yucca.pipeline.preprocessing import UnsupervisedPreprocessor


class YuccaPlanner(object):
    """
    For subclasses changing parameters: remember to also change the name.
    otherwise different planners will save files with identical names and override eachother.

    target_coordinate_system defines the starting orientation for medical images.
    This is only used if the nifti files contain valid headers.

    transpose_forward is applied AFTER images are set to starting orientation defined by
    self.target_coordinate_system, to train on specific views
    (e.g. 2.5D model ensembles where each 2D model is trained on one of the axial, coronal
    and sagittal views).

    transpose_backward is used to revert any transformation applied by transpose_forward.

    crop_to_nonzero removes all the background outside the bounding box of the object
    (e.g. the brain).

    norm_op defines the normalization operation applied in preprocessing.

    Assuming volumes are RAS, i.e. [sagittal, coronal, axial] == [left-right, back-front, down-up],
    this means 2D models will be trained on sagittal slices with this planner.
    """

    def __init__(
        self,
        task,
        preprocessor="YuccaPreprocessor",
        threads=None,
        disable_cc_analysis=True,
        disable_sanity_checks=False,
        view=None,
    ):
        # Required arguments
        self.task = task

        # Planner Specific settings
        self.name = str(self.__class__.__name__) + str(view or "")
        self.compress = False
        self.target_coordinate_system = "RAS"
        self.crop_to_nonzero = True
        self.norm_op = "standardize"
        self.allow_missing_modalities = False
        self.get_foreground_locations_per_label = False

        # This is only relevant for planners with fixed sizes.
        self.keep_aspect_ratio_when_using_target_size = True

        # Don't change the remaining variables unless you know what you're doing
        # Threading speeds up the process. Unittests should by default be enabled.
        self.preprocessor = preprocessor
        self.threads = int(threads) if threads is not None else 1
        self.disable_sanity_checks = disable_sanity_checks
        self.disable_cc_analysis = disable_cc_analysis

        self.plans = {}
        self.suggested_dimensionality = "3D"
        self.view = view

    def plan(self):
        self.set_paths()

        if not isfile(join(self.target_dir, "dataset_properties.pkl")):
            print("Properties file not found. Creating one now - this might take a few minutes")
            create_dataset_properties(data_dir=self.in_dir, save_dir=self.target_dir, num_workers=self.threads)
        self.dataset_properties = load_pickle(join(self.target_dir, "dataset_properties.pkl"))

        self.determine_transpose()
        self.determine_target_size_from_fixed_size_or_spacing()
        self.determine_task_type()
        self.validate_target_size()
        self.drop_keys_from_dict(dict=self.dataset_properties, keys=[])
        self.populate_plans_file()

        save_json(self.plans, self.plans_path, sort_keys=False)

    def determine_transpose(self):
        # If no specific view is determined in run training, we select the optimal.
        # This will be the optimal solution in most cases that are not 2.5D training.
        dims = self.dataset_properties["data_dimensions"]

        if dims == 2:
            self.suggested_dimensionality = "2D"

        if not self.view:
            median_size = np.median(self.dataset_properties["original_sizes"], 0)
            sorting_key = median_size.argsort()
            median_size_sorted = median_size[sorting_key]
            median_spacing_sorted = np.median(self.dataset_properties["original_spacings"], 0)[sorting_key]
            if median_size_sorted[0] < median_size_sorted[1] / 2 and median_spacing_sorted[0] > median_spacing_sorted[1] * 2:
                self.transpose_fw = sorting_key.tolist()
                self.transpose_bw = sorting_key.argsort().tolist()
                self.suggested_dimensionality = "2D"
            else:
                self.transpose_fw = [0, 1, 2][:dims]
                self.transpose_bw = [0, 1, 2][:dims]

        if self.view == "X":
            self.suggested_dimensionality = "2D"
            self.transpose_fw = [0, 1, 2][:dims]
            self.transpose_bw = [0, 1, 2][:dims]

        if self.view == "Y":
            self.suggested_dimensionality = "2D"
            self.transpose_fw = [1, 0, 2][:dims]
            self.transpose_bw = [1, 0, 2][:dims]

        if self.view == "Z":
            self.suggested_dimensionality = "2D"
            assert dims == 3, "Z-view does not exist for 2D data"
            self.transpose_fw = [2, 1, 0]
            self.transpose_bw = [2, 1, 0]

        assert self.transpose_fw is not None, "no transposition, something is wrong."

    def validate_target_size(self):
        assert self.fixed_target_size is None or self.fixed_target_spacing is None, (
            "only one of target size or target spacing should be specified"
            f" but both are specified here as {self.fixed_target_size} and {self.fixed_target_spacing} respectively"
        )

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_size = None
        self.fixed_target_spacing = self.dataset_properties["original_median_spacing"]

    def drop_keys_from_dict(self, dict, keys):
        for key in keys:
            dict.pop(key)

    def preprocess(self):
        preprocessor = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "pipeline", "preprocessing")],
            class_name=self.preprocessor,
            current_module="yucca.pipeline.preprocessing",
        )

        preprocessor = preprocessor(
            plans_path=self.plans_path,
            task=self.task,
            threads=self.threads,
            disable_sanity_checks=self.disable_sanity_checks,
            allow_missing_modalities=self.allow_missing_modalities,
            compress=self.compress,
            get_foreground_locs_per_label=self.get_foreground_locations_per_label,
        )
        preprocessor.run()
        self.postprocess()

    def populate_plans_file(self):
        self.plans = make_plans_file(
            allow_missing_modalities=self.allow_missing_modalities,
            crop_to_nonzero=self.crop_to_nonzero,
            classes=self.dataset_properties["classes"],
            norm_op=self.norm_op,
            modalities=self.dataset_properties["modalities"],
            plans_name=self.name,
            preprocessor=self.preprocessor,
            dataset_properties=self.dataset_properties,
            keep_aspect_ratio_when_using_target_size=self.keep_aspect_ratio_when_using_target_size,
            task_type=self.task_type,
            target_coordinate_system=self.target_coordinate_system,
            target_spacing=self.fixed_target_spacing,
            target_size=self.fixed_target_size,
            transpose_forward=self.transpose_fw,
            transpose_backward=self.transpose_bw,
            suggested_dimensionality=self.suggested_dimensionality,
        )

    def postprocess(self):
        self.plans = add_stats_to_plans_post_preprocessing(plans=self.plans, directory=self.plans_folder)
        save_json(self.plans, self.plans_path, sort_keys=False)

    def set_paths(self):
        # Setting up paths
        self.in_dir = join(yucca_raw_data, self.task)
        self.target_dir = join(yucca_preprocessed_data, self.task)
        self.plans_folder = join(self.target_dir, self.name)
        self.plans_path = join(self.plans_folder, self.name + "_plans.json")
        maybe_mkdir_p(join(self.target_dir, self.name))

    def determine_task_type(self):
        preprocessor_class = recursive_find_python_class(
            folder=[join(yucca.__path__[0], "pipeline", "preprocessing")],
            class_name=self.preprocessor,
            current_module="yucca.pipeline.preprocessing",
        )
        # If key is not present in plan then we try to infer the task_type from the Type of Preprocessor
        assert (
            preprocessor_class
        ), f"{self.preprocessor} was found in plans, but no class with the corresponding name was found"
        if issubclass(preprocessor_class, ClassificationPreprocessor):
            self.task_type = "classification"
        elif issubclass(preprocessor_class, UnsupervisedPreprocessor):
            self.task_type = "self-supervised"
        else:
            self.task_type = "segmentation"


class YuccaPlannerX(YuccaPlanner):
    """
    Used to train (mostly 2D) models on the Coronal view of 3D volumes.
    """

    def __init__(
        self,
        task,
        preprocessor="YuccaPreprocessor",
        threads=12,
        disable_cc_analysis=True,
        disable_sanity_checks=False,
        view=None,
    ):
        super().__init__(
            task,
            preprocessor=preprocessor,
            threads=threads,
            disable_cc_analysis=disable_cc_analysis,
            disable_sanity_checks=disable_sanity_checks,
            view=view,
        )
        self.view = "X"


class YuccaPlannerY(YuccaPlanner):
    """
    Used to train (mostly 2D) models on the Coronal view of 3D volumes.
    """

    def __init__(
        self,
        task,
        preprocessor="YuccaPreprocessor",
        threads=None,
        disable_cc_analysis=True,
        disable_sanity_checks=False,
        view=None,
    ):
        super().__init__(
            task,
            preprocessor=preprocessor,
            threads=threads,
            disable_cc_analysis=disable_cc_analysis,
            disable_sanity_checks=disable_sanity_checks,
            view=view,
        )
        self.view = "Y"


class YuccaPlannerZ(YuccaPlanner):
    """
    Used to train (mostly 2D) models on the Axial view of 3D volumes.
    """

    def __init__(
        self,
        task,
        preprocessor="YuccaPreprocessor",
        threads=None,
        disable_cc_analysis=True,
        disable_sanity_checks=False,
        view=None,
    ):
        super().__init__(
            task,
            preprocessor=preprocessor,
            threads=threads,
            disable_cc_analysis=disable_cc_analysis,
            disable_sanity_checks=disable_sanity_checks,
            view=view,
        )
        self.view = "Z"


class UnsupervisedPlanner(YuccaPlanner):
    def __init__(self, task, preprocessor=None, threads=None, disable_sanity_checks=False, view=None):
        super().__init__(
            task, preprocessor=preprocessor, threads=threads, disable_sanity_checks=disable_sanity_checks, view=view
        )
        self.name = str(self.__class__.__name__)
        self.norm_op = "volume_wise_znorm"
        self.preprocessor = "UnsupervisedPreprocessor"  # hard coded


class YuccaPlannerMinMax(YuccaPlanner):
    """
    Standardizes the images to 0-1 range.
    """

    def __init__(self, task, preprocessor=None, threads=None, disable_sanity_checks=False, view=None):
        super().__init__(
            task, preprocessor=preprocessor, threads=threads, disable_sanity_checks=disable_sanity_checks, view=view
        )
        self.name = str(self.__class__.__name__)
        self.norm_op = "minmax"


class YuccaPlanner_224x224_MinMax(YuccaPlanner):
    def __init__(self, task, preprocessor="YuccaPreprocessor", threads=None, disable_unittests=False, view=None):
        super().__init__(task, preprocessor, threads, disable_unittests, view)
        self.name = str(self.__class__.__name__) + str(view or "")
        self.norm_op = "255to1"

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_size = (224, 224)
        self.fixed_target_spacing = None



def test_inifite_random_sampler():
    import numpy as np
    from yucca.data.samplers import RandomBatchSampler

    dataset = np.arange(100)
    batch_size = 10
    sampler = RandomBatchSampler(dataset, batch_size)

    # repeat test 5 times
    for i in range(5):
        sample = list(sampler)[0]
        assert len(sample) == batch_size



import shutil
import nibabel as nib
from sklearn.model_selection import train_test_split
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json
from yucca.paths import yucca_raw_data, yucca_source


def convert(path: str = yucca_source, subdir: str = "AIMS-TBI24"):
    # INPUT DATA
    path = f"{path}/{subdir}"

    # Train/Test Splits
    images_dir = labels_dir = path
    samples = [i[: -len("_Lesion.nii.gz")] for i in subfiles(labels_dir, suffix="_Lesion.nii.gz", join=False)]
    training_samples, test_samples = train_test_split(samples, random_state=35219)

    ###OUTPUT DATA
    # Target names
    task_name = "Task015_AIMS-TBI24"
    task_prefix = "TBI24"

    # Target paths
    target_base = join(yucca_raw_data, task_name)

    target_imagesTr = join(target_base, "imagesTr")
    target_labelsTr = join(target_base, "labelsTr")

    target_imagesTs = join(target_base, "imagesTs")
    target_labelsTs = join(target_base, "labelsTs")

    maybe_mkdir_p(target_imagesTr)
    maybe_mkdir_p(target_labelsTs)
    maybe_mkdir_p(target_imagesTs)
    maybe_mkdir_p(target_labelsTr)

    for sTr in training_samples:
        src_image_file_path = join(images_dir, sTr + "_T1.nii.gz")
        src_label_file_path = join(images_dir, sTr + "_Lesion.nii.gz")

        label = nib.load(src_label_file_path)
        labelarr = label.get_fdata()
        labelarr[labelarr > 0] = 1
        label = nib.Nifti1Image(labelarr, label.affine, label.header)

        dst_image_file_path = f"{target_imagesTr}/{task_prefix}_{sTr}_000.nii.gz"
        dst_label_path = f"{target_labelsTr}/{task_prefix}_{sTr}.nii.gz"

        shutil.copy2(src_image_file_path, dst_image_file_path)
        nib.save(label, dst_label_path)

    del sTr
    for sTs in test_samples:
        src_image_file_path = join(images_dir, sTs + "_T1.nii.gz")
        src_label_file_path = join(images_dir, sTs + "_Lesion.nii.gz")

        label = nib.load(src_label_file_path)
        labelarr = label.get_fdata()
        labelarr[labelarr > 0] = 1
        label = nib.Nifti1Image(labelarr, label.affine, label.header)

        dst_image_file_path = f"{target_imagesTs}/{task_prefix}_{sTs}_000.nii.gz"
        dst_label_path = f"{target_labelsTs}/{task_prefix}_{sTs}.nii.gz"

        shutil.copy2(src_image_file_path, dst_image_file_path)
        nib.save(label, dst_label_path)

    del sTs
    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        target_imagesTs,
        modalities=("MRI",),
        labels={
            0: "background",
            1: "Lesion",
        },
        dataset_name=task_name,
        license="CC-BY-SA 4.0",
        dataset_description="MICCAI AIMS-TBI 2024",
        dataset_reference="https://aims-tbi.grand-challenge.org/",
    )


if __name__ == "__main__":
    convert()



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner


class YuccaPlanner_MaxSize(YuccaPlanner):
    def __init__(
        self,
        task,
        preprocessor="YuccaPreprocessor",
        threads=None,
        disable_cc_analysis=True,
        disable_sanity_checks=False,
        view=None,
    ):
        super().__init__(task, preprocessor, threads, disable_cc_analysis, disable_sanity_checks, view)
        self.name = str(self.__class__.__name__) + str(view or "")

    def determine_target_size_from_fixed_size_or_spacing(self):
        self.fixed_target_size = self.dataset_properties["original_max_size"]
        self.fixed_target_spacing = None



import numpy as np
from yucca.data.augmentation.transforms.YuccaTransform import YuccaTransform
from yucca.functional.transforms.croppad import croppad
from typing import Literal, Union


class CropPad(YuccaTransform):
    def __init__(
        self,
        data_key="image",
        label_key="label",
        pad_value: Union[Literal["min", "zero", "edge"], int, float] = "min",
        patch_size: tuple | list = None,
        p_oversample_foreground=0.0,
    ):
        self.data_key = data_key
        self.label_key = label_key
        self.pad_value = pad_value
        self.patch_size = patch_size
        self.p_oversample_foreground = p_oversample_foreground

    @staticmethod
    def get_params(data, pad_value, target_shape):
        if pad_value == "min":
            pad_kwargs = {"constant_values": data.min(), "mode": "constant"}
        elif pad_value == "zero":
            pad_kwargs = {"constant_values": np.zeros(1, dtype=data.dtype), "mode": "constant"}
        elif isinstance(pad_value, int) or isinstance(pad_value, float):
            pad_kwargs = {"constant_values": pad_value, "mode": "constant"}
        elif pad_value == "edge":
            pad_kwargs = {"mode": "edge"}
        else:
            print("Unrecognized pad value detected.")
        input_shape = data.shape
        target_image_shape = (input_shape[0], *target_shape)
        target_label_shape = (1, *target_shape)
        return input_shape, target_image_shape, target_label_shape, pad_kwargs

    def __croppad__(
        self,
        data_dict: np.ndarray,
        image_properties: dict,
        input_shape: np.ndarray,
        p_oversample_foreground: float,
        target_image_shape: list | tuple,
        target_label_shape: list | tuple,
        **pad_kwargs,
    ):
        image = data_dict[self.data_key]
        if data_dict.get(self.label_key) is not None:
            label = data_dict[self.label_key]
        else:
            label = None

        image, label = croppad(
            image=image,
            image_properties=image_properties,
            label=label,
            input_dims=len(input_shape[1:]),
            patch_size=self.patch_size,
            p_oversample_foreground=p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )

        data_dict[self.data_key] = image
        if label is not None:
            data_dict[self.label_key] = label
        return data_dict

    def __call__(self, packed_data_dict=None, image_properties=None, **unpacked_data_dict):
        data_dict = packed_data_dict if packed_data_dict else unpacked_data_dict

        input_shape, target_image_shape, target_label_shape, pad_kwargs = self.get_params(
            data=data_dict[self.data_key], pad_value=self.pad_value, target_shape=self.patch_size
        )

        data_dict = self.__croppad__(
            data_dict=data_dict,
            image_properties=image_properties,
            input_shape=input_shape,
            p_oversample_foreground=self.p_oversample_foreground,
            target_image_shape=target_image_shape,
            target_label_shape=target_label_shape,
            **pad_kwargs,
        )
        return data_dict



from yucca.pipeline.planning.YuccaPlanner import YuccaPlanner


class ClassificationPlanner(YuccaPlanner):
    def __init__(self, task, preprocessor=None, threads=None, disable_unittests=False, view=None):
        super().__init__(task, preprocessor, threads, disable_unittests, view)
        self.name = str(self.__class__.__name__) + str(view or "")
        self.preprocessor = "ClassificationPreprocessor"



import numpy as np


def additive_noise(image, mean, sigma, clip_to_input_range: bool = False):
    # J = I+n
    img_min = image.min()
    img_max = image.max()
    image += np.random.normal(mean, sigma, image.shape)
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image


def multiplicative_noise(image, mean, sigma, clip_to_input_range: bool = False):
    # J = I + I*n
    img_min = image.min()
    img_max = image.max()
    gauss = np.random.normal(mean, sigma, image.shape)
    image += image * gauss
    if clip_to_input_range:
        image = np.clip(image, a_min=img_min, a_max=img_max)
    return image



import nibabel as nib
import numpy as np
import nibabel.orientations as nio
from yucca.functional.utils.nib_utils import get_nib_orientation, get_nib_spacing


def verify_spacing_is_equal(reference: nib.Nifti1Image, target: nib.Nifti1Image, id: str = ""):
    assert np.allclose(get_nib_spacing(reference), get_nib_spacing(target)), (
        f"Spacings do not match for {id}"
        f"Image is: {get_nib_spacing(reference)} while the label is {get_nib_spacing(target)}"
    )


def verify_orientation_is_equal(reference: nib.Nifti1Image, target: nib.Nifti1Image, id: str = ""):
    assert get_nib_orientation(reference) == get_nib_orientation(target), (
        f"Directions do not match for {id}"
        f"Image is: {get_nib_orientation(reference)} while the label is {get_nib_orientation(target)}"
    )


def verify_nifti_header_is_valid(image: nib.Nifti1Image):
    if image.get_qform(coded=True)[1] or image.get_sform(coded=True)[1]:
        return True
    else:
        return False


def verify_orientation_is_LR_PA_IS(image: nib.Nifti1Image):
    """
    Checks whether images are in the RAS/LPI domain, which corresponds to:
    X = Left/Right (left/right)
    Y = Posterior/Anterior (backwards/forwards)
    Z = Inferior/Superior (down/up)

    If this is not the case, then images should be converted to RAS (or at least some combination of LR-PA-IS)
    during task conversion as features such as resampling to a target spacing may become unreliably or incorrect
    """
    expected_orientation_code = np.array([0.0, 1.0, 2.0])  # This means LR-PA-IS
    orientation = get_nib_orientation(image)
    if np.all(nio.axcodes2ornt(orientation)[:, 0] == expected_orientation_code):
        return True
    else:
        return False



from yucca.pipeline.managers.YuccaManager import YuccaManager


class YuccaManager_1epoch(YuccaManager):
    def __init__(
        self,
        model,
        model_dimensions: str,
        task: str,
        folds: str | int,
        plan_id: str,
        starting_lr: float = None,
        loss_fn: str = None,
        momentum: float = None,
        continue_training: bool = False,
        checkpoint: str = None,
        finetune: bool = False,
        fast_training: bool = False,
    ):
        super().__init__(
            model,
            model_dimensions,
            task,
            folds,
            plan_id,
            starting_lr,
            loss_fn,
            momentum,
            continue_training,
            checkpoint,
            finetune,
            fast_training,
        )
        self.final_epoch = 1



from batchgenerators.utilities.file_and_folder_operations import subfiles, join, maybe_mkdir_p
import shutil

"""
Task031_
Task032_
Task033_
Task034_
Task035_
Task036_
Task037_
Task038_
Task039_
Task040_
"""

prefixes = [
    "DC_BrainTumour_",
    "DC_Heart_",
    "DC_Liver_",
    "DC_Hippocampus_",
    "DC_Prostate_",
    "DC_Lung_",
    "DC_Pancreas_",
    "DC_HepaticVessel_",
    "DC_Spleen_",
    "DC_Colon_",
]

expected = [
    "Task01_BrainTumour",
    "Task02_Heart",
    "Task03_Liver",
    "Task04_Hippocampus",
    "Task05_Prostate",
    "Task06_Lung",
    "Task07_Pancreas",
    "Task08_HepaticVessel",
    "Task09_Spleen",
    "Task10_Colon",
]

folders = [
    "/home/zcr545/YuccaData/yucca_predictions/Task031_DC_BrainTumour/Task031_DC_BrainTumour/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task032_DC_Heart/Task032_DC_Heart/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task033_DC_Liver/Task033_DC_Liver/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task034_DC_Hippocampus/Task034_DC_Hippocampus/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task035_DC_Prostate/Task035_DC_Prostate/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task036_DC_Lung/Task036_DC_Lung/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task037_DC_Pancreas/Task037_DC_Pancreas/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task038_DC_HepaticVessel/Task038_DC_HepaticVessel/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task039_DC_Spleen/Task039_DC_Spleen/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
    "/home/zcr545/YuccaData/yucca_predictions/Task040_DC_Colon/Task040_DC_Colon/UNet3D/YuccaTrainerV2__YuccaPlannerV2/fold_0_checkpoint_best",
]

blacklist = [
    "Task03_Liver/liver_141.nii.gz",
    "Task03_Liver/liver_156.nii.gz",
    "Task03_Liver/liver_160.nii.gz",
    "Task03_Liver/liver_161.nii.gz",
    "Task03_Liver/liver_162.nii.gz",
    "Task03_Liver/liver_164.nii.gz",
    "Task03_Liver/liver_167.nii.gz",
    "Task03_Liver/liver_182.nii.gz",
    "Task03_Liver/liver_189.nii.gz",
    "Task03_Liver/liver_190.nii.gz",
    "Task08_HepaticVessel/hepaticvessel_247.nii.gz",
]

savepath = "/home/zcr545/decathlon_submission"
maybe_mkdir_p(savepath)

for i in range(10):
    for seg_path in subfiles(folders[i], join=False):
        new_filename = seg_path[len(prefixes[i]) :]
        new_path = join(expected[i], new_filename)
        if new_path in blacklist:
            print("blocking: ", new_path)
            continue
        maybe_mkdir_p(join(savepath, expected[i]))

        shutil.copy(join(folders[i], seg_path), f"{savepath}/{expected[i]}/{new_filename}")



import torch
import torch.nn as nn
import torch.utils.checkpoint as checkpoint
from yucca.networks.blocks_and_layers.conv_blocks import (
    MedNeXtBlock,
    MedNeXtDownBlock,
    MedNeXtUpBlock,
    OutBlock,
)
from yucca.networks.networks.YuccaNet import YuccaNet


class MedNeXt(YuccaNet):
    """
    From the paper: https://arxiv.org/pdf/2303.09975.pdf
    code source: https://github.com/MIC-DKFZ/MedNeXt/tree/main
    Using the 5-kernel L version presented in the paper.
    """

    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        conv_op=None,
        starting_filters: int = 32,
        exp_r=[3, 4, 8, 8, 8, 8, 8, 4, 3],  # Expansion ratio as in Swin Transformers
        kernel_size: int = 5,  # Ofcourse can test kernel_size
        enc_kernel_size: int = None,
        dec_kernel_size: int = None,
        deep_supervision: bool = False,  # Can be used to test deep supervision
        do_res: bool = True,  # Can be used to individually test residual connection
        do_res_up_down: bool = True,  # Additional 'res' connection on up and down convs
        checkpoint_style="outside_block",  # Either inside block or outside block
        block_counts: list = [
            3,
            4,
            8,
            8,
            8,
            8,
            8,
            4,
            3,
        ],  # Can be used to test staging ratio:
        norm_type="group",
        grn=False,
    ):
        super().__init__()

        self.deep_supervision = deep_supervision
        self.num_classes = num_classes
        assert checkpoint_style in [None, "outside_block"]
        if checkpoint_style == "outside_block":
            self.outside_block_checkpointing = True
        else:
            self.outside_block_checkpointing = False
        if kernel_size is not None:
            enc_kernel_size = kernel_size
            dec_kernel_size = kernel_size

        if conv_op == nn.Conv2d:
            dim = "2d"
        else:
            dim = "3d"

        self.stem = conv_op(input_channels, starting_filters, kernel_size=1)
        if isinstance(exp_r, int):
            exp_r = [exp_r for i in range(len(block_counts))]

        self.enc_block_0 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters,
                    out_channels=starting_filters,
                    exp_r=exp_r[0],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[0])
            ]
        )

        self.down_0 = MedNeXtDownBlock(
            in_channels=starting_filters,
            out_channels=2 * starting_filters,
            exp_r=exp_r[1],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
        )

        self.enc_block_1 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 2,
                    out_channels=starting_filters * 2,
                    exp_r=exp_r[1],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[1])
            ]
        )

        self.down_1 = MedNeXtDownBlock(
            in_channels=2 * starting_filters,
            out_channels=4 * starting_filters,
            exp_r=exp_r[2],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.enc_block_2 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 4,
                    out_channels=starting_filters * 4,
                    exp_r=exp_r[2],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[2])
            ]
        )

        self.down_2 = MedNeXtDownBlock(
            in_channels=4 * starting_filters,
            out_channels=8 * starting_filters,
            exp_r=exp_r[3],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.enc_block_3 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 8,
                    out_channels=starting_filters * 8,
                    exp_r=exp_r[3],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[3])
            ]
        )

        self.down_3 = MedNeXtDownBlock(
            in_channels=8 * starting_filters,
            out_channels=16 * starting_filters,
            exp_r=exp_r[4],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.bottleneck = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 16,
                    out_channels=starting_filters * 16,
                    exp_r=exp_r[4],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[4])
            ]
        )

        self.up_3 = MedNeXtUpBlock(
            in_channels=16 * starting_filters,
            out_channels=8 * starting_filters,
            exp_r=exp_r[5],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_3 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 8,
                    out_channels=starting_filters * 8,
                    exp_r=exp_r[5],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[5])
            ]
        )

        self.up_2 = MedNeXtUpBlock(
            in_channels=8 * starting_filters,
            out_channels=4 * starting_filters,
            exp_r=exp_r[6],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_2 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 4,
                    out_channels=starting_filters * 4,
                    exp_r=exp_r[6],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[6])
            ]
        )

        self.up_1 = MedNeXtUpBlock(
            in_channels=4 * starting_filters,
            out_channels=2 * starting_filters,
            exp_r=exp_r[7],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_1 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 2,
                    out_channels=starting_filters * 2,
                    exp_r=exp_r[7],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[7])
            ]
        )

        self.up_0 = MedNeXtUpBlock(
            in_channels=2 * starting_filters,
            out_channels=starting_filters,
            exp_r=exp_r[8],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_0 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters,
                    out_channels=starting_filters,
                    exp_r=exp_r[8],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[8])
            ]
        )

        self.out_0 = OutBlock(in_channels=starting_filters, n_classes=self.num_classes, dim=dim)

        # Used to fix PyTorch checkpointing bug
        self.dummy_tensor = nn.Parameter(torch.tensor([1.0]), requires_grad=True)

        if self.deep_supervision:
            self.out_1 = OutBlock(in_channels=starting_filters * 2, n_classes=self.num_classes, dim=dim)
            self.out_2 = OutBlock(in_channels=starting_filters * 4, n_classes=self.num_classes, dim=dim)
            self.out_3 = OutBlock(in_channels=starting_filters * 8, n_classes=self.num_classes, dim=dim)
            self.out_4 = OutBlock(in_channels=starting_filters * 16, n_classes=self.num_classes, dim=dim)

        self.block_counts = block_counts

    def iterative_checkpoint(self, sequential_block, x):
        """
        This simply forwards x through each block of the sequential_block while
        using gradient_checkpointing. This implementation is designed to bypass
        the following issue in PyTorch's gradient checkpointing:
        https://discuss.pytorch.org/t/checkpoint-with-no-grad-requiring-inputs-problem/19117/9
        """
        for l in sequential_block:
            x = checkpoint.checkpoint(l, x, self.dummy_tensor)
        return x

    def forward(self, x):
        x = self.stem(x)
        if self.outside_block_checkpointing:
            x_res_0 = self.iterative_checkpoint(self.enc_block_0, x)
            x = checkpoint.checkpoint(self.down_0, x_res_0, self.dummy_tensor)
            x_res_1 = self.iterative_checkpoint(self.enc_block_1, x)
            x = checkpoint.checkpoint(self.down_1, x_res_1, self.dummy_tensor)
            x_res_2 = self.iterative_checkpoint(self.enc_block_2, x)
            x = checkpoint.checkpoint(self.down_2, x_res_2, self.dummy_tensor)
            x_res_3 = self.iterative_checkpoint(self.enc_block_3, x)
            x = checkpoint.checkpoint(self.down_3, x_res_3, self.dummy_tensor)

            x = self.iterative_checkpoint(self.bottleneck, x)
            if self.deep_supervision:
                x_ds_4 = checkpoint.checkpoint(self.out_4, x, self.dummy_tensor)

            x_up_3 = checkpoint.checkpoint(self.up_3, x, self.dummy_tensor)
            dec_x = x_res_3 + x_up_3
            x = self.iterative_checkpoint(self.dec_block_3, dec_x)
            if self.deep_supervision:
                x_ds_3 = checkpoint.checkpoint(self.out_3, x, self.dummy_tensor)
            del x_res_3, x_up_3

            x_up_2 = checkpoint.checkpoint(self.up_2, x, self.dummy_tensor)
            dec_x = x_res_2 + x_up_2
            x = self.iterative_checkpoint(self.dec_block_2, dec_x)
            if self.deep_supervision:
                x_ds_2 = checkpoint.checkpoint(self.out_2, x, self.dummy_tensor)
            del x_res_2, x_up_2

            x_up_1 = checkpoint.checkpoint(self.up_1, x, self.dummy_tensor)
            dec_x = x_res_1 + x_up_1
            x = self.iterative_checkpoint(self.dec_block_1, dec_x)
            if self.deep_supervision:
                x_ds_1 = checkpoint.checkpoint(self.out_1, x, self.dummy_tensor)
            del x_res_1, x_up_1

            x_up_0 = checkpoint.checkpoint(self.up_0, x, self.dummy_tensor)
            dec_x = x_res_0 + x_up_0
            x = self.iterative_checkpoint(self.dec_block_0, dec_x)
            del x_res_0, x_up_0, dec_x

            x = checkpoint.checkpoint(self.out_0, x, self.dummy_tensor)

        else:
            x_res_0 = self.enc_block_0(x)
            x = self.down_0(x_res_0)
            x_res_1 = self.enc_block_1(x)
            x = self.down_1(x_res_1)
            x_res_2 = self.enc_block_2(x)
            x = self.down_2(x_res_2)
            x_res_3 = self.enc_block_3(x)
            x = self.down_3(x_res_3)

            x = self.bottleneck(x)
            if self.deep_supervision:
                x_ds_4 = self.out_4(x)

            x_up_3 = self.up_3(x)
            dec_x = x_res_3 + x_up_3
            x = self.dec_block_3(dec_x)

            if self.deep_supervision:
                x_ds_3 = self.out_3(x)
            del x_res_3, x_up_3

            x_up_2 = self.up_2(x)
            dec_x = x_res_2 + x_up_2
            x = self.dec_block_2(dec_x)
            if self.deep_supervision:
                x_ds_2 = self.out_2(x)
            del x_res_2, x_up_2

            x_up_1 = self.up_1(x)
            dec_x = x_res_1 + x_up_1
            x = self.dec_block_1(dec_x)
            if self.deep_supervision:
                x_ds_1 = self.out_1(x)
            del x_res_1, x_up_1

            x_up_0 = self.up_0(x)
            dec_x = x_res_0 + x_up_0
            x = self.dec_block_0(dec_x)
            del x_res_0, x_up_0, dec_x

            x = self.out_0(x)

        if self.deep_supervision:
            return [x, x_ds_1, x_ds_2, x_ds_3, x_ds_4]
        else:
            return x


class MedNeXt_Modular(YuccaNet):
    """
    From the paper: https://arxiv.org/pdf/2303.09975.pdf
    code source: https://github.com/MIC-DKFZ/MedNeXt/tree/main
    Using the 5-kernel L version presented in the paper.
    """

    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        conv_op=None,
        starting_filters: int = 32,
        exp_r=[3, 4, 8, 8, 8, 8, 8, 4, 3],  # Expansion ratio as in Swin Transformers
        kernel_size: int = 5,  # Ofcourse can test kernel_size
        enc_kernel_size: int = None,
        dec_kernel_size: int = None,
        deep_supervision: bool = False,  # Can be used to test deep supervision
        do_res: bool = True,  # Can be used to individually test residual connection
        do_res_up_down: bool = True,  # Additional 'res' connection on up and down convs
        checkpoint_style="outside_block",  # Either inside block or outside block
        block_counts: list = [
            3,
            4,
            8,
            8,
            8,
            8,
            8,
            4,
            3,
        ],  # Can be used to test staging ratio:
        norm_type="group",
        grn=False,
    ):
        super().__init__()

        self.encoder = MedNeXt_encoder(
            input_channels=input_channels,
            num_classes=num_classes,
            conv_op=conv_op,
            starting_filters=starting_filters,
            exp_r=exp_r,
            kernel_size=kernel_size,
            enc_kernel_size=enc_kernel_size,
            dec_kernel_size=dec_kernel_size,
            deep_supervision=deep_supervision,
            do_res=do_res,
            do_res_up_down=do_res_up_down,
            checkpoint_style=checkpoint_style,
            block_counts=block_counts,
            norm_type=norm_type,
            grn=grn,
        )
        self.decoder = MedNeXt_decoder(
            input_channels=input_channels,
            num_classes=num_classes,
            conv_op=conv_op,
            starting_filters=starting_filters,
            kernel_size=kernel_size,
            dec_kernel_size=dec_kernel_size,
            deep_supervision=deep_supervision,
            do_res=do_res,
            do_res_up_down=do_res_up_down,
            checkpoint_style=checkpoint_style,
            block_counts=block_counts,
            norm_type=norm_type,
            grn=grn,
        )

    def forward(self, x):
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MedNeXt_encoder(nn.Module):
    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        conv_op=None,
        starting_filters: int = 32,
        exp_r=[3, 4, 8, 8, 8, 8, 8, 4, 3],  # Expansion ratio as in Swin Transformers
        kernel_size: int = 5,  # Ofcourse can test kernel_size
        enc_kernel_size: int = None,
        dec_kernel_size: int = None,
        deep_supervision: bool = False,  # Can be used to test deep supervision
        do_res: bool = True,  # Can be used to individually test residual connection
        do_res_up_down: bool = True,  # Additional 'res' connection on up and down convs
        checkpoint_style="outside_block",  # Either inside block or outside block
        block_counts: list = [
            3,
            4,
            8,
            8,
            8,
            8,
            8,
            4,
            3,
        ],  # Can be used to test staging ratio:
        norm_type="group",
        grn=False,
    ):
        super().__init__()

        self.deep_supervision = deep_supervision
        self.num_classes = num_classes
        assert checkpoint_style in [None, "outside_block"]
        if checkpoint_style == "outside_block":
            self.outside_block_checkpointing = True
        else:
            self.outside_block_checkpointing = False
        if kernel_size is not None:
            enc_kernel_size = kernel_size
            dec_kernel_size = kernel_size

        if conv_op == nn.Conv2d:
            dim = "2d"
        else:
            dim = "3d"

        self.stem = conv_op(input_channels, starting_filters, kernel_size=1)
        if isinstance(exp_r, int):
            exp_r = [exp_r for i in range(len(block_counts))]

        self.enc_block_0 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters,
                    out_channels=starting_filters,
                    exp_r=exp_r[0],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[0])
            ]
        )

        self.down_0 = MedNeXtDownBlock(
            in_channels=starting_filters,
            out_channels=2 * starting_filters,
            exp_r=exp_r[1],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
        )

        self.enc_block_1 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 2,
                    out_channels=starting_filters * 2,
                    exp_r=exp_r[1],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[1])
            ]
        )

        self.down_1 = MedNeXtDownBlock(
            in_channels=2 * starting_filters,
            out_channels=4 * starting_filters,
            exp_r=exp_r[2],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.enc_block_2 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 4,
                    out_channels=starting_filters * 4,
                    exp_r=exp_r[2],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[2])
            ]
        )

        self.down_2 = MedNeXtDownBlock(
            in_channels=4 * starting_filters,
            out_channels=8 * starting_filters,
            exp_r=exp_r[3],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.enc_block_3 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 8,
                    out_channels=starting_filters * 8,
                    exp_r=exp_r[3],
                    kernel_size=enc_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[3])
            ]
        )

        self.down_3 = MedNeXtDownBlock(
            in_channels=8 * starting_filters,
            out_channels=16 * starting_filters,
            exp_r=exp_r[4],
            kernel_size=enc_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.bottleneck = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 16,
                    out_channels=starting_filters * 16,
                    exp_r=exp_r[4],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[4])
            ]
        )

        # Used to fix PyTorch checkpointing bug
        self.dummy_tensor = nn.Parameter(torch.tensor([1.0]), requires_grad=True)

    def iterative_checkpoint(self, sequential_block, x):
        for l in sequential_block:
            x = checkpoint.checkpoint(l, x, self.dummy_tensor)
        return x

    def forward(self, x):
        x = self.stem(x)
        if self.outside_block_checkpointing:
            x_res_0 = self.iterative_checkpoint(self.enc_block_0, x)
            x = checkpoint.checkpoint(self.down_0, x_res_0, self.dummy_tensor)
            x_res_1 = self.iterative_checkpoint(self.enc_block_1, x)
            x = checkpoint.checkpoint(self.down_1, x_res_1, self.dummy_tensor)
            x_res_2 = self.iterative_checkpoint(self.enc_block_2, x)
            x = checkpoint.checkpoint(self.down_2, x_res_2, self.dummy_tensor)
            x_res_3 = self.iterative_checkpoint(self.enc_block_3, x)
            x = checkpoint.checkpoint(self.down_3, x_res_3, self.dummy_tensor)
            x = self.iterative_checkpoint(self.bottleneck, x)
        else:
            x_res_0 = self.enc_block_0(x)
            x = self.down_0(x_res_0)
            x_res_1 = self.enc_block_1(x)
            x = self.down_1(x_res_1)
            x_res_2 = self.enc_block_2(x)
            x = self.down_2(x_res_2)
            x_res_3 = self.enc_block_3(x)
            x = self.down_3(x_res_3)

            x = self.bottleneck(x)
        return [x_res_0, x_res_1, x_res_2, x_res_3, x]


class MedNeXt_decoder(nn.Module):
    def __init__(
        self,
        input_channels: int,
        num_classes: int = 1,
        conv_op=None,
        starting_filters: int = 32,
        exp_r=[3, 4, 8, 8, 8, 8, 8, 4, 3],  # Expansion ratio as in Swin Transformers
        kernel_size: int = 5,  # Ofcourse can test kernel_size
        dec_kernel_size: int = None,
        deep_supervision: bool = False,  # Can be used to test deep supervision
        do_res: bool = True,  # Can be used to individually test residual connection
        do_res_up_down: bool = True,  # Additional 'res' connection on up and down convs
        checkpoint_style="outside_block",  # Either inside block or outside block
        block_counts: list = [
            3,
            4,
            8,
            8,
            8,
            8,
            8,
            4,
            3,
        ],  # Can be used to test staging ratio:
        norm_type="group",
        grn=False,
    ):
        super().__init__()

        self.deep_supervision = deep_supervision
        self.num_classes = num_classes
        assert checkpoint_style in [None, "outside_block"]
        if checkpoint_style == "outside_block":
            self.outside_block_checkpointing = True
        else:
            self.outside_block_checkpointing = False
        if kernel_size is not None:
            dec_kernel_size = kernel_size

        if conv_op == nn.Conv2d:
            dim = "2d"
        else:
            dim = "3d"

        self.stem = conv_op(input_channels, starting_filters, kernel_size=1)
        if isinstance(exp_r, int):
            exp_r = [exp_r for i in range(len(block_counts))]

        self.up_3 = MedNeXtUpBlock(
            in_channels=16 * starting_filters,
            out_channels=8 * starting_filters,
            exp_r=exp_r[5],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_3 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 8,
                    out_channels=starting_filters * 8,
                    exp_r=exp_r[5],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[5])
            ]
        )

        self.up_2 = MedNeXtUpBlock(
            in_channels=8 * starting_filters,
            out_channels=4 * starting_filters,
            exp_r=exp_r[6],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_2 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 4,
                    out_channels=starting_filters * 4,
                    exp_r=exp_r[6],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[6])
            ]
        )

        self.up_1 = MedNeXtUpBlock(
            in_channels=4 * starting_filters,
            out_channels=2 * starting_filters,
            exp_r=exp_r[7],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_1 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters * 2,
                    out_channels=starting_filters * 2,
                    exp_r=exp_r[7],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[7])
            ]
        )

        self.up_0 = MedNeXtUpBlock(
            in_channels=2 * starting_filters,
            out_channels=starting_filters,
            exp_r=exp_r[8],
            kernel_size=dec_kernel_size,
            do_res=do_res_up_down,
            norm_type=norm_type,
            dim=dim,
            grn=grn,
        )

        self.dec_block_0 = nn.Sequential(
            *[
                MedNeXtBlock(
                    in_channels=starting_filters,
                    out_channels=starting_filters,
                    exp_r=exp_r[8],
                    kernel_size=dec_kernel_size,
                    do_res=do_res,
                    norm_type=norm_type,
                    dim=dim,
                    grn=grn,
                )
                for i in range(block_counts[8])
            ]
        )

        self.out_0 = OutBlock(in_channels=starting_filters, n_classes=self.num_classes, dim=dim)

        # Used to fix PyTorch checkpointing bug
        self.dummy_tensor = nn.Parameter(torch.tensor([1.0]), requires_grad=True)

        if self.deep_supervision:
            raise NotImplementedError

        self.block_counts = block_counts

    def iterative_checkpoint(self, sequential_block, x):
        for l in sequential_block:
            x = checkpoint.checkpoint(l, x, self.dummy_tensor)
        return x

    def forward(self, xs: list):
        # unpack the output of the encoder
        x_res_0, x_res_1, x_res_2, x_res_3, x = xs
        if self.outside_block_checkpointing:
            x_up_3 = checkpoint.checkpoint(self.up_3, x, self.dummy_tensor)
            dec_x = x_res_3 + x_up_3
            x = self.iterative_checkpoint(self.dec_block_3, dec_x)

            del x_res_3, x_up_3

            x_up_2 = checkpoint.checkpoint(self.up_2, x, self.dummy_tensor)
            dec_x = x_res_2 + x_up_2
            x = self.iterative_checkpoint(self.dec_block_2, dec_x)

            del x_res_2, x_up_2

            x_up_1 = checkpoint.checkpoint(self.up_1, x, self.dummy_tensor)
            dec_x = x_res_1 + x_up_1
            x = self.iterative_checkpoint(self.dec_block_1, dec_x)

            del x_res_1, x_up_1

            x_up_0 = checkpoint.checkpoint(self.up_0, x, self.dummy_tensor)
            dec_x = x_res_0 + x_up_0
            x = self.iterative_checkpoint(self.dec_block_0, dec_x)
            del x_res_0, x_up_0, dec_x

            x = checkpoint.checkpoint(self.out_0, x, self.dummy_tensor)

        else:
            x_up_3 = self.up_3(x)
            dec_x = x_res_3 + x_up_3
            x = self.dec_block_3(dec_x)

            del x_res_3, x_up_3

            x_up_2 = self.up_2(x)
            dec_x = x_res_2 + x_up_2
            x = self.dec_block_2(dec_x)
            del x_res_2, x_up_2

            x_up_1 = self.up_1(x)
            dec_x = x_res_1 + x_up_1
            x = self.dec_block_1(dec_x)

            del x_res_1, x_up_1

            x_up_0 = self.up_0(x)
            dec_x = x_res_0 + x_up_0
            x = self.dec_block_0(dec_x)
            del x_res_0, x_up_0, dec_x

            x = self.out_0(x)

        return x



import numpy as np


def crop_to_box(array, bbox):
    """
    Crops an array to the Bounding Box indices
    Should be a list of [xmin, xmax, ymin, ymax (, zmin, zmax)]

    We add +1 because slicing excludes the high val index. Which it should not do here.
    """
    if len(bbox) > 5:
        bbox_slices = (
            slice(bbox[0], bbox[1] + 1),
            slice(bbox[2], bbox[3] + 1),
            slice(bbox[4], bbox[5] + 1),
        )
    else:
        bbox_slices = (slice(bbox[0], bbox[1] + 1), slice(bbox[2], bbox[3] + 1))
    return array[bbox_slices]


def pad_to_size(array, size, **kwargs):
    pad_box = get_pad_box(array, size)
    if len(pad_box) > 5:
        array_padded = np.pad(
            array,
            (
                (pad_box[0], pad_box[1]),
                (pad_box[2], pad_box[3]),
                (pad_box[4], pad_box[5]),
            ),
            **kwargs,
        )
        return array_padded, pad_box

    array_padded = np.pad(
        array,
        ((pad_box[0], pad_box[1]), (pad_box[2], pad_box[3])),
        **kwargs,
    )
    return array_padded, pad_box


def get_pad_box(original_array, min_size):
    assert len(original_array.shape) in [2, 3], "incorrect array shape"
    assert len(min_size) in [2, 3], "incorrect patch_size shape"
    pad_box = []

    if len(original_array.shape) == 3:
        if len(min_size) == 2:
            # 3D Data for 2D model.
            min_size = (0, *min_size)
        for i in range(3):
            val = max(0, min_size[i] - original_array.shape[i])
            pad_box.append(val // 2)
            pad_box.append(val // 2 + val % 2)
        return pad_box

    # 2D data + 2D model
    for i in range(2):
        val = max(0, min_size[i] - original_array.shape[i])
        pad_box.append(val // 2)
        pad_box.append(val // 2 + val % 2)
    return pad_box


def get_pad_kwargs(data, pad_value):
    if pad_value == "min":
        pad_kwargs = {"constant_values": data.min(), "mode": "constant"}
    elif pad_value == "zero":
        pad_kwargs = {"constant_values": np.zeros(1, dtype=data.dtype), "mode": "constant"}
    elif isinstance(pad_value, int) or isinstance(pad_value, float):
        pad_kwargs = {"constant_values": pad_value, "mode": "constant"}
    elif pad_value == "edge":
        pad_kwargs = {"mode": "edge"}
    else:
        print("Unrecognized pad value detected.")
    return pad_kwargs



import numpy as np


def remove_small_objects(array: np.ndarray, min_size: float | int, voxel_spacing: np.ndarray | list):
    """
    This function will remove any connected components smaller
    than the minimum size in mm
    """
    labels, counts = np.unique(array, return_counts=True)
    labels = list(labels)

    # Remove background
    labels.remove(0)

    voxel_size = np.prod(voxel_spacing)

    for label in labels:
        if not voxel_size * counts[label] >= min_size:
            array[array == label] = 0
            labels.remove(label)

    return array, labels



import numpy as np
from batchgenerators.utilities.file_and_folder_operations import isfile
from yucca.data.datasets.YuccaDataset import YuccaTrainDataset


class YuccaCompressedTrainDataset(YuccaTrainDataset):
    def load_and_maybe_keep_volume(self, path: str):
        path = path + ".npz"
        if isfile(path):
            try:
                return np.load(path, "r")["data"]
            except ValueError:
                return np.load(path, allow_pickle=True)["data"]
        else:
            print("compressed data not found.")



import shutil
from batchgenerators.utilities.file_and_folder_operations import join, maybe_mkdir_p, subfiles
from yucca.pipeline.task_conversion.utils import generate_dataset_json, should_use_volume, dirs_in_dir
from yucca.paths import yucca_raw_data
from tqdm import tqdm
import nibabel as nib


rejec_modalities = ["fmri", "func"]  # TODO: Might need to skip GRE


def should_skip(modality: str):
    modality = modality.lower()
    for reject in rejec_modalities:
        if reject in modality:
            return True
    return False


def convert(path: str, subdir: str = "OASIS4"):
    """INPUT DATA - Define input path and suffixes"""
    path = join(path, subdir)

    """ OUTPUT DATA - Define the task name and prefix """
    task_name = "Task204_OASIS4"
    task_prefix = "OASIS4"

    subjects_dir = join(path, "OASIS4_images")
    target_base = join(yucca_raw_data, task_name)
    target_imagesTr = join(target_base, "imagesTr")

    maybe_mkdir_p(target_imagesTr)

    """Populate Target Directory"""

    ext = ".nii.gz"

    skipped_volumes = []
    skipped_modalities = []
    for subject in tqdm(dirs_in_dir(subjects_dir), desc="Subjects"):
        if "mr" not in subject.lower():
            continue  # skip this subject

        subject_dir = join(subjects_dir, subject)
        for modality in dirs_in_dir(subject_dir):
            if should_skip(modality):
                skipped_modalities.append(modality)
                continue  # skip this modality
            modality_dir = join(subject_dir, modality, "NIFTI")
            for file in subfiles(modality_dir, join=False, suffix=ext):
                image_path = join(modality_dir, file)
                other_info = file[: -len(ext)].replace(subject, "").replace(modality, "")
                vol = nib.load(image_path)
                if should_use_volume(vol):
                    output_name = f"{task_prefix}_{subject}_{modality}_{other_info}_000.nii.gz"
                    print(output_name)
                    output_path = join(target_imagesTr, output_name)
                    shutil.copy2(image_path, output_path)
                else:
                    skipped_volumes.append(image_path)

    print(f"Skipped {len(skipped_volumes)} and {len(skipped_modalities)}")
    print("Skipped workers", skipped_volumes)
    print("Skipped modalities", skipped_modalities)

    generate_dataset_json(
        join(target_base, "dataset.json"),
        target_imagesTr,
        imagesTs_dir=None,
        modalities=["MRI"],
        labels={},
        dataset_name=task_name,
        license="https://www.oasis-brains.org/#access",
        dataset_description="Open Access Series of Imaging Studies (OASIS)",
        dataset_reference="https://www.oasis-brains.org",
    )



import numpy as np
import os
import shutil
import nibabel as nib
from yucca.paths import yucca_raw_data
from typing import Literal
from batchgenerators.utilities.file_and_folder_operations import save_json, subfiles, join, subdirs
from tqdm import tqdm
from pathlib import Path


def combine_images_from_tasks(tasks: list, target_base: str, run_type: Literal["supervised", "unsupervised"]):
    assert len(tasks) > 0, "list of tasks empty"
    for task in tqdm(tasks):
        folders = ["imagesTr", "imagesTs", "labelsTr", "labelsTs"] if run_type == "supervised" else ["imagesTr"]
        for folder in folders:
            source = os.path.join(yucca_raw_data, task, folder)
            target = os.path.join(target_base, folder)
            print("Copying ", source, target)
            copy_files_from_to(source, target)


def copy_files_from_to(source_dir, target_dir):
    assert os.path.isdir(source_dir)
    os.makedirs(target_dir, exist_ok=True)

    for file in tqdm(os.listdir(source_dir)):
        shutil.copy2(os.path.join(source_dir, file), f"{target_dir}/{file}")


def get_identifiers_from_splitted_files(folder: str, ext, tasks: list):
    if len(tasks) > 0:
        uniques = np.unique(
            [i[: -len("_000." + ext)] for task in tasks for i in subfiles(join(folder, task, suffix=ext), join=False)]
        )
    else:
        uniques = np.unique([i[: -len("_000." + ext)] for i in subfiles(folder, suffix=ext, join=False)])
    return list(uniques)


def dirs_in_dir(dir: str):
    p = Path(dir)
    return [f.name for f in p.iterdir() if f.is_dir() and f.name[0] not in [".", "_"]]


def files_in_dir(dir: str):
    p = Path(dir)
    return [f.name for f in p.iterdir() if f.is_file() and f.name[0] not in [".", "_"]]


def should_use_volume(vol: nib.Nifti1Image):
    return not (np.any(np.array(vol.shape) < 15) or len(vol.shape) != 3 or np.array(vol.dataobj).min() < 0)


def remove_punctuation_and_spaces(data: str):
    data = data.replace(" ", "-").replace(",", "-").replace(".", "-")
    return data


def generate_dataset_json(
    output_file: str,
    imagesTr_dir: str,
    imagesTs_dir: str,
    modalities: dict,
    labels: dict,
    dataset_name: str,
    label_hierarchy: dict = {},
    regions_in_order=[],  # We do not want to do these as dicts, as its sensitive to order and they easily become sorted by mistake
    regions_labeled=[],
    tasks: list = [],
    license: str = "hands off!",
    dataset_description: str = "",
    dataset_reference="",
    dataset_release="0.0",
):
    """
    :param output_file: This needs to be the full path to the dataset.json you intend to write, so
    output_file='DATASET_PATH/dataset.json' where the folder DATASET_PATH points to is the one with the
    imagesTr and labelsTr subfolders
    :param imagesTr_dir: path to the imagesTr folder of that dataset
    :param imagesTs_dir: path to the imagesTs folder of that dataset. Can be None
    :param modalities: dict of modality names and their corresponding values. must be in the same order as the images (first entry
    corresponds to _000.nii.gz, etc). Example: ('T1', 'T2', 'FLAIR').
    :param labels: dict with int->str (key->value) mapping the label IDs to label names. Note that 0 is always
    supposed to be background! Example: {0: 'background', 1: 'left hippocampus', 2: 'right hippocampus'}
    :param dataset_name: The name of the dataset. Can be anything you want
    :param license:
    :param dataset_description:
    :param dataset_reference: website of the dataset, if available
    :param dataset_release:
    :return:
    """
    first_file = files_in_dir(imagesTr_dir)[0]
    im_ext = os.path.split(first_file)[-1].split(os.extsep, 1)[-1]
    train_identifiers = get_identifiers_from_splitted_files(imagesTr_dir, im_ext, tasks)

    if imagesTs_dir is not None:
        test_identifiers = get_identifiers_from_splitted_files(imagesTs_dir, im_ext, tasks)
    else:
        test_identifiers = []

    json_dict = {}
    json_dict["name"] = dataset_name
    json_dict["description"] = dataset_description
    json_dict["tensorImageSize"] = "4D"
    json_dict["reference"] = dataset_reference
    json_dict["licence"] = license
    json_dict["release"] = dataset_release
    json_dict["image_extension"] = im_ext
    json_dict["modality"] = {str(i): modalities[i] for i in range(len(modalities))}
    json_dict["labels"] = {str(i): labels[i] for i in labels.keys()} if labels is not None else None
    json_dict["label_hierarchy"] = label_hierarchy
    json_dict["regions_in_order"] = regions_in_order
    json_dict["regions_labeled"] = regions_labeled
    json_dict["tasks"] = tasks
    json_dict["numTraining"] = len(train_identifiers)
    json_dict["numTest"] = len(test_identifiers)
    json_dict["training"] = [{"image": name, "label": name if labels else None} for name in train_identifiers]
    json_dict["test"] = test_identifiers

    if not output_file.endswith("dataset.json"):
        print(
            "WARNING: output file name is not dataset.json! This may be intentional or not. You decide. "
            "Proceeding anyways..."
        )
    save_json(json_dict, os.path.join(output_file))


def maybe_get_task_from_task_id(task_id: str | int):
    task_id = str(task_id)
    tasks = subdirs(yucca_raw_data, join=False)

    # Check if name is already complete
    if task_id in tasks:
        return task_id

    # If not, we try to recreate the name
    # We use the raw_data folder as reference
    for task in tasks:
        if task_id.lower() in task.lower():
            return task

    # If we can't find anything we just return the original, on the offchance that the task does not exist in Raw Data while existing in e.g. Preprocessed
    print(
        f"Couldn't find a task called: {task_id} in the raw data folder: {yucca_raw_data}. If your task only exists in e.g. the Preprocessed folder things might still work."
    )
    return task_id



from yucca.pipeline.managers.YuccaManager import YuccaManager


class YuccaManager_Tester(YuccaManager):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.augmentation_params = {
            "mirror_p_per_sample": 0.0,
            "mirror_p_per_axis": 0.0,
            "motion_ghosting_p_per_sample": 0.0,
            "multiplicative_noise_p_per_sample": 0.0,
            "rotation_p_per_sample": 0.0,
            "rotation_p_per_axis": 0.0,
            "scale_p_per_sample": 0.0,
            "simulate_lowres_p_per_sample": 0.0,
            "simulate_lowres_p_per_channel": 0.0,
            "simulate_lowres_p_per_axis": 0.0,
            "normalize": True,
            "normalization_scheme": "minmax",
        }



__author__ = "Ezequiel de la Rosa"
import numpy as np
import warnings
import cc3d
import argparse
from batchgenerators.utilities.file_and_folder_operations import subfiles, join, save_json


def compute_dice(im1, im2, empty_value=1.0):
    """
    Computes the Dice coefficient, a measure of set similarity.
    Parameters
    ----------
    im1 : array-like, bool
        Any array of arbitrary size. If not boolean, will be converted.
    im2 : array-like, bool
        Any other array of identical size as im1. If not boolean, it will be converted.
    empty_value : scalar, float.

    Returns
    -------
    dice : float
        Dice coefficient as a float on range [0,1].
        Maximum similarity = 1
        No similarity = 0
        If both images are empty (sum equal to zero) = empty_value

    Notes
    -----
    The order of inputs for `dice` is irrelevant. The result will be
    identical if `im1` and `im2` are switched.

    This function has been adapted from the Verse Challenge repository:
    https://github.com/anjany/verse/blob/main/utils/eval_utilities.py
    """

    im1 = np.asarray(im1).astype(bool)
    im2 = np.asarray(im2).astype(bool)

    if im1.shape != im2.shape:
        raise ValueError("Shape mismatch: im1 and im2 must have the same shape.")

    im_sum = im1.sum() + im2.sum()
    if im_sum == 0:
        return empty_value

    # Compute Dice coefficient
    intersection = np.logical_and(im1, im2)

    return 2.0 * intersection.sum() / im_sum


def compute_absolute_volume_difference(im1, im2, voxel_size):
    """
    Computes the absolute volume difference between two masks.

    Parameters
    ----------
    im1 : array-like, bool
        Any array of arbitrary size. If not boolean, will be converted.
    im2 : array-like, bool
        Any other array of identical size as 'ground_truth'. If not boolean, it will be converted.
    voxel_size : scalar, float (ml)
        If not float, it will be converted.

    Returns
    -------
    abs_vol_diff : float, measured in ml.
        Absolute volume difference as a float.
        Maximum similarity = 0
        No similarity = inf


    Notes
    -----
    The order of inputs is irrelevant. The result will be identical if `im1` and `im2` are switched.
    """

    im1 = np.asarray(im1).astype(bool)
    im2 = np.asarray(im2).astype(bool)
    voxel_size = voxel_size.astype(np.float)

    if im1.shape != im2.shape:
        warnings.warn(
            "Shape mismatch: ground_truth and prediction have difference shapes."
            " The absolute volume difference is computed with mismatching shape masks"
        )

    ground_truth_volume = np.sum(im1) * voxel_size
    prediction_volume = np.sum(im2) * voxel_size
    abs_vol_diff = np.abs(ground_truth_volume - prediction_volume)

    return abs_vol_diff


def compute_absolute_lesion_difference(ground_truth, prediction, connectivity=26):
    """
    Computes the absolute lesion difference between two masks. The number of lesions are counted for
    each volume, and their absolute difference is computed.

    Parameters
    ----------
    ground_truth : array-like, bool
        Any array of arbitrary size. If not boolean, will be converted.
    prediction : array-like, bool
        Any other array of identical size as 'ground_truth'. If not boolean, it will be converted.

    Returns
    -------
    abs_les_diff : int
        Absolute lesion difference as integer.
        Maximum similarity = 0
        No similarity = inf


    Notes
    -----
    """
    ground_truth = np.asarray(ground_truth).astype(bool)
    prediction = np.asarray(prediction).astype(bool)

    _, ground_truth_numb_lesion = cc3d.connected_components(ground_truth, connectivity=connectivity, return_N=True)
    _, prediction_numb_lesion = cc3d.connected_components(prediction, connectivity=connectivity, return_N=True)
    abs_les_diff = abs(ground_truth_numb_lesion - prediction_numb_lesion)
    print(ground_truth_numb_lesion, prediction_numb_lesion)

    return abs_les_diff


def compute_lesion_f1_score(ground_truth, prediction, empty_value=1.0, connectivity=26):
    """
    Computes the lesion-wise F1-score between two masks.

    Parameters
    ----------
    ground_truth : array-like, bool
        Any array of arbitrary size. If not boolean, will be converted.
    prediction : array-like, bool
        Any other array of identical size as 'ground_truth'. If not boolean, it will be converted.
    empty_value : scalar, float.
    connectivity : scalar, int.

    Returns
    -------
    f1_score : float
        Lesion-wise F1-score as float.
        Max score = 1
        Min score = 0
        If both images are empty (tp + fp + fn =0) = empty_value

    Notes
    -----
    This function computes lesion-wise score by defining true positive lesions (tp), false positive lesions (fp) and
    false negative lesions (fn) using 3D connected-component-analysis.

    tp: 3D connected-component from the ground-truth image that overlaps at least on one voxel with the prediction image.
    fp: 3D connected-component from the prediction image that has no voxel overlapping with the ground-truth image.
    fn: 3d connected-component from the ground-truth image that has no voxel overlapping with the prediction image.
    """
    ground_truth = np.asarray(ground_truth).astype(bool)
    prediction = np.asarray(prediction).astype(bool)
    tp = 0
    fp = 0
    fn = 0

    # Check if ground-truth connected-components are detected or missed (tp and fn respectively).
    intersection = np.logical_and(ground_truth, prediction)
    labeled_ground_truth, N = cc3d.connected_components(ground_truth, connectivity=connectivity, return_N=True)

    # Iterate over ground_truth clusters to find tp and fn.
    # tp and fn are only computed if the ground-truth is not empty.
    if N > 0:
        for _, binary_cluster_image in cc3d.each(labeled_ground_truth, binary=True, in_place=True):
            if np.logical_and(binary_cluster_image, intersection).any():
                tp += 1
            else:
                fn += 1

    # iterate over prediction clusters to find fp.
    # fp are only computed if the prediction image is not empty.
    labeled_prediction, N = cc3d.connected_components(prediction, connectivity=connectivity, return_N=True)
    if N > 0:
        for _, binary_cluster_image in cc3d.each(labeled_prediction, binary=True, in_place=True):
            if not np.logical_and(binary_cluster_image, ground_truth).any():
                fp += 1

    # Define case when both images are empty.
    if tp + fp + fn == 0:
        _, N = cc3d.connected_components(ground_truth, connectivity=connectivity, return_N=True)
        if N == 0:
            f1_score = empty_value
    else:
        f1_score = tp / (tp + (fp + fn) / 2)

    return f1_score


import nibabel as nib

gts = [
    "/home/zcr545/YuccaData/yucca_raw_data/Task049_3BrainLesion3Labels/labelsTs/3BL3L_1.nii.gz",
    "/home/zcr545/YuccaData/yucca_raw_data/Task049_3BrainLesion3Labels/labelsTs/3BL3L_sub-strokecase0049.nii.gz",
    "/home/zcr545/YuccaData/yucca_raw_data/Task049_3BrainLesion3Labels/labelsTs/3BL3L_UNC_train_Case09.nii.gz",
]
preds = [
    "/home/zcr545/YuccaData/yucca_segmentations/Task049_3BrainLesion3Labels/Task049_3BrainLesion3Labels/UNet2D/YuccaTrainerV2__YuccaPlannerV2_Ensemble/fold_0_checkpoint_best/3BL3L_1.nii.gz",
    "/home/zcr545/YuccaData/yucca_segmentations/Task049_3BrainLesion3Labels/Task049_3BrainLesion3Labels/UNet2D/YuccaTrainerV2__YuccaPlannerV2_Ensemble/fold_0_checkpoint_best/3BL3L_sub-strokecase0049.nii.gz",
    "/home/zcr545/YuccaData/yucca_segmentations/Task049_3BrainLesion3Labels/Task049_3BrainLesion3Labels/UNet2D/YuccaTrainerV2__YuccaPlannerV2_Ensemble/fold_0_checkpoint_best/3BL3L_UNC_train_Case09.nii.gz",
]
fp = 0
tp = 0
fn = 0
tplog = []
fnlog = []
labelarr = np.array([1, 2, 3])
for i in range(1):
    gt = nib.load(gts[i]).get_fdata()
    pred = nib.load(preds[i]).get_fdata()
    for label in labelarr:
        pred_for_label = np.where(pred == label, pred, 0).astype(bool)
        gt_for_label = np.where(gt == label, gt, 0).astype(bool)
        intersection = np.logical_and(gt_for_label, pred_for_label)
        gt_cc, gt_cc_n = cc3d.connected_components(gt_for_label, connectivity=26, return_N=True)
        pred_cc, pred_cc_n = cc3d.connected_components(pred_for_label, connectivity=26, return_N=True)
        # Compute Dice coefficient
        if gt_cc_n > 0:
            for _, binary_cluster_image in cc3d.each(gt_cc, binary=True, in_place=True):
                if np.logical_and(binary_cluster_image, pred_cc).any():
                    tp += 1
                    tplog.append(binary_cluster_image.sum())
                else:
                    fn += 1
                    fnlog.append(binary_cluster_image.sum())
        if pred_cc_n > 0:
            for _, binary_cluster_image in cc3d.each(pred_cc, binary=True, in_place=True):
                if not np.logical_and(binary_cluster_image, gt_cc).any():
                    fp += 1
print(tp, fp, fn, gt_cc_n, pred_cc_n)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--pred", required=True)
    parser.add_argument("--gt", required=True)
    args = parser.parse_args()

    pred = args.pred
    gt = args.gt

    predfiles = subfiles(pred, join=False, suffix=".nii.gz")
    metrics = {}
    all_dice = []
    all_abs = []
    all_f1 = []
    for sub in predfiles:
        metrics[sub] = {}
        im1 = nib.load(join(gt, sub)).get_fdata()
        im2 = nib.load(join(pred, sub)).get_fdata()

        dice = compute_dice(im1, im2)
        abs_les_diff = compute_absolute_lesion_difference(im1, im2)
        les_f1 = compute_lesion_f1_score(im1, im2)
        metrics[sub]["dice"] = dice
        all_dice.append(dice)
        metrics[sub]["absolute lesion difference"] = abs_les_diff
        all_abs.append(abs_les_diff)
        metrics[sub]["lesion f1 score"] = les_f1
        all_f1.append(les_f1)

    metrics["mean"] = {}
    metrics["mean"]["dice"] = float(np.mean(all_dice))
    metrics["mean"]["absolute lesion difference"] = float(np.mean(all_abs))
    metrics["mean"]["lesion f1 score"] = float(np.mean(all_f1))
    save_json(metrics, join(pred, "results_isles_eval.json"))



from yucca.functional.array_operations.bounding_boxes import get_bbox_for_foreground
from yucca.functional.array_operations.cropping_and_padding import crop_to_box, get_pad_box, pad_to_size, get_pad_kwargs
from yucca.functional.array_operations.filtering import remove_small_objects
from yucca.functional.array_operations.matrix_ops import (
    create_zero_centered_coordinate_matrix,
    deform_coordinate_matrix,
    Rx,
    Ry,
    Rz,
    Rz2D,
    get_max_rotated_size,
)
from yucca.functional.array_operations.normalization import normalizer, clamp, znormalize, rescale
from yucca.functional.array_operations.transpose import transpose_array, transpose_case



import logging
import math
from typing import Union
import numpy as np
from batchgenerators.utilities.file_and_folder_operations import join, isfile, save_pickle, load_pickle
from sklearn.model_selection import KFold
from dataclasses import dataclass
from yucca.functional.utils.files_and_folders import subfiles
from yucca.pipeline.configuration.configure_paths import PathConfig


@dataclass
class SplitConfig:
    splits: Union[dict[dict[list[dict]]], None] = None  # Contains `{ method: { parameter_value: [splits] }}`
    method: str = None
    param: Union[int, float] = None

    def split(self):
        return self.splits[str(self.method)][self.param]

    def train(self, idx):
        return self.split()[idx]["train"]

    def val(self, idx):
        return self.split()[idx]["val"]

    def lm_hparams(self):
        return {"split_method": self.method, "split_param": self.param}


def get_split_config(method: str, param: Union[str, float, int], path_config: PathConfig):
    """
    Params:
        method: SplitMethods
        param: Int or float depending on method param
    """
    splits_path = join(path_config.task_dir, "splits.pkl")

    if isfile(splits_path):
        splits = load_pickle(splits_path)

        # Overwrite old splits
        if not isinstance(splits, dict):
            splits = {}

        if split_is_precomputed(splits, method, param):
            logging.warning(
                f"Reusing already computed split file which was split using the {method} method and parameter {param}."
            )
            return SplitConfig(splits, method, param)
        else:
            logging.warning("Generating new split since splits did not contain a split computed with the same parameters.")
    else:
        splits = {}

    if method not in splits.keys():
        splits[method] = {}

    file_names = get_file_names(path_config.train_data_dir)

    if method == "kfold":
        splits[method][param] = kfold_split(file_names, int(param))
    elif method == "simple_train_val_split":
        splits[method][param] = simple_split(file_names, float(param))
    else:
        assert split_is_precomputed(
            splits, method, param
        ), "you are using a custom split method, but the split.pkl does not contain such a split"

    split_cfg = SplitConfig(splits, method, param)
    save_pickle(splits, splits_path)
    return split_cfg


def split_is_precomputed(splits, method, param):
    return method in splits.keys() and param in splits[method].keys()


def kfold_split(file_names: list[str], k: int):
    assert k is not None
    kf = KFold(n_splits=k, shuffle=True)
    splits = []
    for train, val in kf.split(file_names):
        splits.append({"train": list(file_names[train]), "val": list(file_names[val])})
    return splits


def simple_split(file_names: list[str], p: float):
    assert p is not None
    assert 0 < p < 1
    np.random.shuffle(file_names)  # inplace
    num_val = math.ceil(len(file_names) * p)
    if num_val < 10:
        logging.warning("The validation split is very small. Consider using a higher `p`.")
    return [{"train": list(file_names[num_val:]), "val": list(file_names[:num_val])}]


def get_file_names(train_data_dir):
    file_names = subfiles(train_data_dir, join=False, remove_extension=True, suffix=".npy")
    if not file_names:
        file_names = subfiles(train_data_dir, join=False, remove_extension=True, suffix=".npz")
        if file_names:
            logging.warning("Only found compressed (.npz) files. This might increase runtime.")

    assert file_names, f"Couldn't find any .npy or .npz files in {train_data_dir}"
    return np.array(file_names)
